"""
file-like wrapper for an open file
"""
from __future__ import absolute_import

import functools
import io

import attr

from .exceptions import ResponseError
from . import ntstatus
from . import smb2


@attr.s
class _Open(object):
    """private

    Base class for ``Open``.

    Handle the SMB2-specific protocol interaction for leases, oplocks, and
    durable handles.

    Wrap :py:class:`~pike.model.Channel` operations which accept an ``Open`` as an
    argument.

    This class shouldn't be instantiated directly. See :py:class:`~pike.io.Open`
    """

    tree = attr.ib()
    create_request = attr.ib()
    create_response = attr.ib()
    create_guid = attr.ib()
    _previous_open = attr.ib()
    oplock_level = attr.ib(init=False)
    lease = attr.ib(init=False)
    _is_resilient = attr.ib(init=False)
    _is_durable = attr.ib(init=False)

    @oplock_level.default
    def _init_oplock_level(self):
        oplock_level = self.create_response.oplock_level
        if oplock_level not in (
            smb2.SMB2_OPLOCK_LEVEL_NONE,
            smb2.SMB2_OPLOCK_LEVEL_LEASE,
        ):
            self.arm_oplock_future()
        return oplock_level

    @lease.default
    def _init_lease(self):
        if self.oplock_level == smb2.SMB2_OPLOCK_LEVEL_LEASE:
            for ctx in self.create_response:
                if isinstance(ctx, smb2.LeaseResponse):
                    return self.tree.session.client.lease(self.tree, ctx)

    @_is_durable.default
    def _init_is_durable(self):
        if self._previous_open is not None:
            return self._previous_open.is_durable
        return (
            self.durable_handle_ctx is not None
            or self.durable_handle_v2_ctx is not None
        )

    @_is_resilient.default
    def _init_is_resilient(self):
        if self._previous_open is not None:
            return self._previous_open.is_resilient
        return False

    @property
    def channel(self):
        """
        :rtype: pike.model.Channel
        :return: The first Channel associated with this handle.

        (The session may have multiple bound channels).
        """
        if not self.closed:
            return self.tree.session.first_channel()
        raise ValueError("I/O operation on closed file")

    @property
    def file_id(self):
        """
        :rtype: 2-tuple of (int, int)
        :return: the server-opaque file handle returned in the CREATE response
        """
        return self.create_response.file_id

    @property
    def durable_handle_ctx(self):
        """
        :rtype: smb2.DurableHandleResponse or None
        """

        for ctx in self.create_response:
            if isinstance(ctx, smb2.DurableHandleResponse):
                return ctx

    @property
    def durable_handle_v2_ctx(self):
        """
        :rtype: smb2.DurableHandleV2Response or None
        """
        for ctx in self.create_response:
            if isinstance(ctx, smb2.DurableHandleV2Response):
                return ctx

    @property
    def durable_timeout(self):
        """
        :rtype: int or None
        :return: the durable handle timeout in seconds (or None if not durable)
        """
        ctx = self.durable_handle_v2_ctx
        if ctx:
            return ctx.durable_timeout
        elif self._previous_open is not None:
            return self._previous_open.durable_timeout

    @property
    def durable_flags(self):
        """
        :rtype: int or None
        :return: the durable handle flags (or None if not durable)
        """
        ctx = self.durable_handle_v2_ctx
        if ctx:
            return ctx.durable_flags
        elif self._previous_open is not None:
            return self._previous_open.durable_flags
        return 0

    @property
    def is_durable(self):
        """
        :return: True if the handle is durable
        """
        return self._is_durable

    @property
    def is_resilient(self):
        """
        :return: True if the handle is resilient
        """
        return self._is_resilient

    @property
    def is_persistent(self):
        """
        :return: True if the handle is persistent
        """
        return self.durable_flags & smb2.SMB2_DHANDLE_FLAG_PERSISTENT != 0

    @property
    def closed(self):
        """
        :return: True if the handle is closed
        """
        return self.tree is None

    def arm_oplock_future(self):
        """
        (Re)arm the oplock future for this open. This function should be called
        when an oplock changes level to anything except SMB2_OPLOCK_LEVEL_NONE
        """
        self.oplock_future = self.tree.session.client.oplock_break_future(self.file_id)

    def on_oplock_break(self, cb):
        """
        Simple oplock break callback handler.

        :type cb: Callable[[pike.smb2.OplockLevel], pike.smb2.OplockLevel])
        :param cb: callback function is passed the
            ``pike.smb2.OplockBreakNotification.oplock_level`` and must return
            the desired :py:class:`pike.smb2.OplockLevel` to break to.
        """

        def simple_handle_break(op, smb_res, cb_ctx):
            """
            note that op is not used in this callback,
            since it already closes over self
            """
            notify = smb_res[0]
            if self.oplock_level != smb2.SMB2_OPLOCK_LEVEL_II:
                ack = self.channel.oplock_break_acknowledgement(self, smb_res)
                ack.oplock_level = cb(notify.oplock_level)
                ack_res = self.channel.connection.transceive(ack.parent.parent)[0][0]
                if ack.oplock_level != smb2.SMB2_OPLOCK_LEVEL_NONE:
                    self.arm_oplock_future()
                    self.on_oplock_break(cb)
                self.oplock_level = ack_res.oplock_level
            else:
                self.oplock_level = notify.oplock_level

        self.on_oplock_break_request(simple_handle_break)

    def on_oplock_break_request(self, cb, cb_ctx=None):
        """
        Complex oplock break callback handler.

        :type cb: Callable[[Open, pike.smb2.Smb2, Optional[Any]], Any])
        :param cb: callback function is passed the file handle, the Smb2
            :py:func:`~pike.smb2.OplockBreakNotification`, and an arbitrary
            context and should handle breaking the oplock in some way. The
            callback is also responsible for calling
            :py:func:`arm_oplock_future` and and updating the ``.oplock_level``
            (if changed)
        :type cb_ctx: Optional[Any]
        :param cb_ctx: arbitrary context passed to the callback
        """

        def handle_break(f):
            smb_res = f.result()
            cb(self, smb_res, cb_ctx)

        self.oplock_future.then(handle_break)

    def dispose(self):
        """
        Clear resources held by this object.

        Called implicitly when the object is closed.
        """
        self.tree = None
        if self.lease is not None:
            self.lease.dispose()
            self.lease = None

    def close(self):
        """
        Close the handle on the server.

        If the channel or connection is unavailable, this is a no-op.
        """
        if self.closed:
            return
        try:
            self.channel.close(self)
            self.dispose()
        except StopIteration:
            # If the underlying connection for the channel is closed explicitly
            # open will not able to find an appropriate channel, to send close.
            pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    # Helper operations call through to self.channel
    @property
    def query_directory_request(self):
        return functools.partial(self.channel.query_directory_request, self)

    @property
    def query_directory(self):
        return functools.partial(self.channel.query_directory, self)

    @property
    def enum_directory(self):
        return functools.partial(self.channel.enum_directory, self)

    @property
    def query_file_info_request(self):
        return functools.partial(self.channel.query_file_info_request, self)

    @property
    def query_file_info(self):
        return functools.partial(self.channel.query_file_info, self)

    @property
    def set_file_info_request(self):
        return functools.partial(self.channel.set_file_info_request, self)

    @property
    def set_file_info(self):
        return functools.partial(self.channel.set_file_info, self)

    @property
    def change_notify_request(self):
        return functools.partial(self.channel.change_notify_request, self)

    @property
    def change_notify(self):
        return functools.partial(self.channel.change_notify, self)

    @property
    def flush_request(self):
        return functools.partial(self.channel.flush_request, self)

    @property
    def flush(self):
        return functools.partial(self.channel.flush, self)

    @property
    def lock_request(self):
        return functools.partial(self.channel.lock_request, self)

    @property
    def lock(self):
        return functools.partial(self.channel.lock, self)

    @property
    def set_symlink_request(self):
        return functools.partial(self.channel.set_symlink_request, self)

    @property
    def set_symlink(self):
        return functools.partial(self.channel.set_symlink, self)

    @property
    def get_symlink_request(self):
        return functools.partial(self.channel.get_symlink_request, self)

    @property
    def get_symlink(self):
        return functools.partial(self.channel.get_symlink, self)

    @property
    def request(self):
        return functools.partial(self.channel.request, obj=self)


class CompatOpen(_Open):
    """
    pike.model.Open compatible __init__ signature.

    Please refactor code to use the new _Open constructor
    """

    def __init__(self, tree, smb_res, create_guid=None, prev=None):
        super(CompatOpen, self).__init__(
            tree=tree,
            create_response=smb_res[0],
            create_guid=create_guid,
            previous_open=prev,
        )


@attr.s
class Open(_Open, io.RawIOBase):
    """
    Represents an open file handle on a remote SMB2 server.

    Can be used as a contextmanager, which will automatically close
    the file when the context block exits.

    This class should not be instantiated directly. Use one of the following
    methods to get an instance:

        * :py:func:`pike.model.Channel.create`
        * :py:func:`pike.path.PikePath.open`

    This class provides a file-like interface with support for ``read``,
    ``write``, ``seek``, ``flush``, and derivatives provided by ``RawIOBase``.

    :vartype tree: pike.model.Tree
    :ivar tree: the :py:class:`~pike.model.Tree` associated with this handle
    :vartype create_request: pike.smb2.CreateRequest
    :ivar create_request: the raw request sent to open this handle
    :vartype create_response: pike.smb2.CreateResponse
    :ivar create_response: the raw response returned by the server
    :vartype create_guid: str
    :ivar create_guid: the create_guid associated with the durable handle
    :vartype oplock_level: pike.smb2.OplockLevel
    :ivar oplock_level: the current Oplock level of the handle
    :vartype lease: pike.model.Lease
    :ivar lease: the current Lease state of the handle
    """

    _offset = attr.ib(default=0, init=False)
    _end_of_file = attr.ib(init=False)

    @_end_of_file.default
    def _init_end_of_file(self):
        return self.create_response.end_of_file

    @property
    def end_of_file(self):
        """
        :return: Byte offset of the end of file as reported by the CREATE response

        Note: this value is only updated by ``write`` and ``read``!
              if other operations are performed directly or via other sessions
              this value may get out of sync.
        """
        return self._end_of_file

    def seek(self, offset, whence=io.SEEK_SET):
        """
        Change the stream position to the given byte offset.

        offset is interpreted relative to the position indicated by whence.

        The default value for whence is SEEK_SET. Values for whence are:

            * ``io.SEEK_SET`` or ``0`` - start of the stream (the default);
              offset should be zero or positive
            * ``io.SEEK_CUR`` or ``1`` - current stream position; offset may be negative
            * ``io.SEEK_END`` or ``2`` - end of the stream; offset is usually negative

        :type offset: int
        :type whence: int (see above)
        :rtype: int
        :return: the new absolute position
        """
        if whence == io.SEEK_SET:
            self._offset = offset
        elif whence == io.SEEK_CUR:
            self._offset += offset
        elif whence == io.SEEK_END:
            self._offset = self.end_of_file + offset
        return self._offset

    def seekable(self):
        """
        :return: True if the stream supports random access.

        All ``Open`` instances support random access.
        """
        return True

    def tell(self):
        """
        :rtype: int
        :return: the absolute byte position within the file
        """
        return self._offset

    def truncate(self, size=None):
        """
        Resize the stream to the given size. NOT SUPPORTED.
        """
        raise NotImplementedError("truncate() not supported")

    def _has_access(self, access):
        if self.create_request is not None:
            return self.create_request.desired_access & access
        return None

    def readable(self):
        """
        :return: True if the stream can be read from
        """
        return any(
            self._has_access(a)
            for a in (smb2.FILE_READ_DATA, smb2.GENERIC_READ, smb2.GENERIC_ALL)
        )

    def _read_range(self, start=0, end=None):
        """
        Read a range of bytes from the file.

        Multiple requests will be made if necessary.

        :type start: int
        :param start: the beginning offset to read from
        :type end: int or None
        :param end: the end offset to read from. If None, read until
            STATUS_END_OF_FILE is returned.
        :rtype: bytes
        :return: bytes read from the file
        """
        max_read_size = self.channel.connection.negotiate_response.max_read_size
        offset = start
        response_buffers = []
        while end is None or offset < end:
            if end is not None:
                max_read_size = min(end - offset, max_read_size)
            available = min(
                self.channel.connection.credits * smb2.BYTES_PER_CREDIT,
                max_read_size,
            )
            try:
                read_resp = self.channel.read(self, available, offset)
                response_buffers.append(read_resp)
                offset += len(read_resp)
            except ResponseError as re:
                if re.response.status == ntstatus.STATUS_END_OF_FILE:
                    break
                raise
        read_buffer = b"".join(rb.tobytes() for rb in response_buffers)
        if read_buffer:
            self._offset = start + len(read_buffer)
            # update the EOF marker if we read past it
            self._end_of_file = max(self.end_of_file, self._offset)
        return read_buffer

    def readall(self):
        """
        Read all bytes from the current offset to EOF.
        """
        return self._read_range(self._offset)

    def readinto(self, b):
        """
        Read bytes into the preallocated buffer, b.

        Note: this function performs extra copying and is inefficient.
        """
        read_resp = self._read_range(self._offset, self._offset + len(b))
        bytes_read = len(read_resp)
        b[:bytes_read] = read_resp
        return bytes_read

    def _write_at(self, data, offset):
        """
        Write bytes to the file at the given offset.

        Multiple requests will be made if necessary.

        :type offset: int
        :param offset: the beginning offset to read from
        :rtype: int
        :return: count of bytes written
        """
        n_data = len(data)
        bytes_written = 0
        while bytes_written < n_data:
            available = min(
                n_data - bytes_written,
                self.channel.connection.credits * smb2.BYTES_PER_CREDIT,
                self.channel.connection.negotiate_response.max_write_size,
            )
            chunk = data[bytes_written : bytes_written + available]
            count = self.channel.write(self, offset + bytes_written, chunk)
            bytes_written += count
        if bytes_written:
            self._offset += bytes_written
            # update the EOF marker if we write past it
            self._end_of_file = max(self.end_of_file, self._offset)
        return bytes_written

    def writable(self):
        """
        :return: True if the stream can be written to
        """
        return any(
            self._has_access(a)
            for a in (
                smb2.FILE_WRITE_DATA,
                smb2.FILE_APPEND_DATA,
                smb2.GENERIC_WRITE,
                smb2.GENERIC_ALL,
            )
        )

    def write(self, b):
        """
        Write bytes from the buffer, b.

        Note: this function may perform extra copying and could be inefficient.
        """
        return self._write_at(b, self._offset)

    def flush(self):
        """
        Issue an SMB2 FLUSH request to the server.

        This instructs the server to flush any intermediate buffers to disk.

        The server may return a response before the data has been flushed.

        For non-writable streams, this call is a no-op.
        """
        if not self.writable():
            # don't flush non-writable streams
            return
        return super(Open, self).flush()



#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        encryption.py
#
# Abstract:
#
#        Test SMB3 Encryption and negotiation options
#
# Authors: Masen Furer (masen.furer@dell.com)
#

import pike.crypto as crypto
import pike.model as model
import pike.smb2 as smb2
import pike.test


class TestEncryption(pike.test.PikeTest):
    def test_smb_3_0_encryption(self):
        client = model.Client(dialects=[smb2.DIALECT_SMB3_0])
        conn = client.connect(self.server)
        conn.negotiate()
        self.assertEqual(conn.negotiate_response.dialect_revision, smb2.DIALECT_SMB3_0)
        self.assertTrue(
            conn.negotiate_response.capabilities & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
        )
        chan = conn.session_setup(self.creds)
        chan.session.encrypt_data = True
        self.assertIsNotNone(chan.session.encryption_context)
        self.assertEqual(chan.session.encryption_context.aes_mode, crypto.AES.MODE_CCM)
        tree = chan.tree_connect(self.share)
        self.assertIsNotNone(tree.tree_connect_response.parent.parent.transform)

    def test_smb_3_0_2_encryption(self):
        client = model.Client(dialects=[smb2.DIALECT_SMB3_0_2])
        conn = client.connect(self.server)
        conn.negotiate()
        self.assertEqual(
            conn.negotiate_response.dialect_revision, smb2.DIALECT_SMB3_0_2
        )
        self.assertTrue(
            conn.negotiate_response.capabilities & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
        )
        chan = conn.session_setup(self.creds)
        chan.session.encrypt_data = True
        self.assertIsNotNone(chan.session.encryption_context)
        self.assertEqual(chan.session.encryption_context.aes_mode, crypto.AES.MODE_CCM)
        tree = chan.tree_connect(self.share)
        self.assertIsNotNone(tree.tree_connect_response.parent.parent.transform)

    def test_smb_3_1_1_encryption_gcm(self):
        client = model.Client(dialects=[smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1])
        conn = client.connect(self.server)
        conn.negotiate(ciphers=[crypto.SMB2_AES_128_GCM])
        self.assertEqual(
            conn.negotiate_response.dialect_revision, smb2.DIALECT_SMB3_1_1
        )
        self.assertFalse(
            conn.negotiate_response.capabilities & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
        )
        chan = conn.session_setup(self.creds)
        chan.session.encrypt_data = True
        self.assertIsNotNone(chan.session.encryption_context)
        self.assertEqual(chan.session.encryption_context.aes_mode, crypto.AES.MODE_GCM)
        tree = chan.tree_connect(self.share)
        self.assertIsNotNone(tree.tree_connect_response.parent.parent.transform)

    def test_smb_3_1_1_encryption_ccm(self):
        client = model.Client(dialects=[smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1])
        conn = client.connect(self.server)
        conn.negotiate(ciphers=[crypto.SMB2_AES_128_CCM])
        self.assertEqual(
            conn.negotiate_response.dialect_revision, smb2.DIALECT_SMB3_1_1
        )
        self.assertFalse(
            conn.negotiate_response.capabilities & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
        )
        chan = conn.session_setup(self.creds)
        chan.session.encrypt_data = True
        self.assertIsNotNone(chan.session.encryption_context)
        self.assertEqual(chan.session.encryption_context.aes_mode, crypto.AES.MODE_CCM)
        tree = chan.tree_connect(self.share)
        self.assertIsNotNone(tree.tree_connect_response.parent.parent.transform)

    def test_smb_3_1_1_compound(self):
        client = model.Client(dialects=[smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1])
        conn = client.connect(self.server)
        conn.negotiate(ciphers=[crypto.SMB2_AES_128_GCM])
        self.assertEqual(
            conn.negotiate_response.dialect_revision, smb2.DIALECT_SMB3_1_1
        )
        self.assertFalse(
            conn.negotiate_response.capabilities & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
        )
        chan = conn.session_setup(self.creds)
        chan.session.encrypt_data = True
        self.assertIsNotNone(chan.session.encryption_context)
        self.assertEqual(chan.session.encryption_context.aes_mode, crypto.AES.MODE_GCM)
        chan.session.encrypt_data = True
        tree = chan.tree_connect(self.share)
        self.assertIsNotNone(tree.tree_connect_response.parent.parent.transform)

        nb_req = chan.frame()
        smb_req1 = chan.request(nb_req, obj=tree)
        smb_req2 = chan.request(nb_req, obj=tree)
        create_req = smb2.CreateRequest(smb_req1)
        close_req = smb2.CloseRequest(smb_req2)

        create_req.name = "hello.txt"
        create_req.desired_access = pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE
        create_req.file_attributes = pike.smb2.FILE_ATTRIBUTE_NORMAL
        create_req.create_disposition = pike.smb2.FILE_OPEN_IF

        max_req = pike.smb2.MaximalAccessRequest(create_req)

        close_req.file_id = smb2.RELATED_FID
        smb_req2.flags |= smb2.SMB2_FLAGS_RELATED_OPERATIONS
        resp = chan.connection.transceive(nb_req)
        parent = resp[0].parent
        self.assertIsNotNone(parent.transform)
        for r in resp:
            self.assertEqual(r.parent, parent)


if __name__ == "__main__":
    pike.test.unittest.main()



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        oplock.py
#
# Abstract:
#
#        Oplock tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.model
import pike.smb2
import pike.test


class OplockTest(pike.test.PikeTest):
    # Open a handle with an oplock and break it
    def test_oplock_break(self):
        chan, tree = self.tree_connect()

        share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )

        handle1 = chan.create(
            tree,
            "oplock.txt",
            share=share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_EXCLUSIVE,
        ).result()

        handle1.on_oplock_break(lambda level: level)

        handle2 = chan.create(
            tree,
            "oplock.txt",
            share=share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_II,
        ).result()

        chan.close(handle1)
        chan.close(handle2)



#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#

import six

if six.PY2:
    import array

    oldarray = array.array
    # monkey-patch tobytes for python 2.7
    class newarray(oldarray):
        def tobytes(self, *args, **kwargs):
            return self.tostring(*args, **kwargs)

        def __getslice__(self, *args, **kwargs):
            return type(self)(
                self.typecode, super(newarray, self).__getslice__(*args, **kwargs)
            )

        def __add__(self, *args, **kwargs):
            return type(self)(
                self.typecode, super(newarray, self).__add__(*args, **kwargs)
            )

    array.array = newarray

# main entry point
from pike.test import TreeConnect as TreeConnect

__all__ = [
    "auth",
    "core",
    "crypto",
    "digest",
    "io",
    "kerberos",
    "model",
    "netbios",
    "ntlm",
    "nttime",
    "ntstatus",
    "path",
    "smb2",
    "test",
    "transport",
    "TreeConnect",
]
# __version__ is defined by setuptools_scm using git tag
# https://github.com/pypa/setuptools_scm/



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        compound.py
#
# Abstract:
#
#        Compound request tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.model
import pike.ntstatus
import pike.smb2
import pike.test


RelatedOpen = pike.model.RelatedOpen


class CompoundTest(pike.test.PikeTest):

    # Compounded create/close of the same file, with maximal access request
    def test_create_close(self):
        chan, tree = self.tree_connect()

        # Manually assemble a chained request
        nb_req = chan.frame()
        smb_req1 = chan.request(nb_req, obj=tree)
        smb_req2 = chan.request(nb_req, obj=tree)
        create_req = pike.smb2.CreateRequest(smb_req1)
        close_req = pike.smb2.CloseRequest(smb_req2)

        create_req.name = "hello.txt"
        create_req.desired_access = pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE
        create_req.file_attributes = pike.smb2.FILE_ATTRIBUTE_NORMAL
        create_req.create_disposition = pike.smb2.FILE_OPEN_IF

        max_req = pike.smb2.MaximalAccessRequest(create_req)

        close_req.file_id = pike.smb2.RELATED_FID
        smb_req2.flags |= pike.smb2.SMB2_FLAGS_RELATED_OPERATIONS

        chan.connection.transceive(nb_req)

    # Compound create/query/close
    def test_create_query_close(self):
        chan, tree = self.tree_connect()

        create_req = chan.create_request(
            tree,
            "create_query_close",
            access=(
                pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE | pike.smb2.DELETE
            ),
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        )
        nb_req = create_req.parent.parent
        query_req = chan.query_file_info_request(RelatedOpen(tree))
        nb_req.adopt(query_req.parent)

        close_req = chan.close_request(RelatedOpen(tree))
        nb_req.adopt(close_req.parent)

        (create_res, query_res, close_res) = chan.connection.transceive(nb_req)

        compare_attributes = [
            "creation_time",
            "change_time",
            "last_access_time",
            "last_write_time",
            "file_attributes",
        ]
        for attr in compare_attributes:
            self.assertEqual(
                getattr(create_res[0], attr), getattr(query_res[0][0], attr)
            )

    # Compound create/write/close & create/read/close
    def test_create_write_close(self):
        filename = "create_write_close"
        buf = b"compounded write"

        chan, tree = self.tree_connect()

        create_req1 = chan.create_request(
            tree, filename, access=pike.smb2.GENERIC_WRITE
        )
        nb_req = create_req1.parent.parent
        write_req = chan.write_request(RelatedOpen(tree), 0, buf)
        nb_req.adopt(write_req.parent)

        close_req = chan.close_request(RelatedOpen(tree))
        nb_req.adopt(close_req.parent)

        (create_res1, write_res, close_res) = chan.connection.transceive(nb_req)

        create_req2 = chan.create_request(
            tree,
            filename,
            access=(pike.smb2.GENERIC_READ | pike.smb2.DELETE),
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        )
        nb_req = create_req2.parent.parent
        read_req = chan.read_request(RelatedOpen(tree), 1024, 0)
        nb_req.adopt(read_req.parent)

        close_req = chan.close_request(RelatedOpen(tree))
        nb_req.adopt(close_req.parent)

        (create_res2, read_res, close_res) = chan.connection.transceive(nb_req)

        self.assertEqual(buf, read_res[0].data.tobytes())

    # Compound create/write/close with insufficient access
    def test_create_write_close_access_denied(self):
        filename = "create_write_close_access_denied"

        chan, tree = self.tree_connect()

        create_req1 = chan.create_request(
            tree,
            filename,
            access=(pike.smb2.GENERIC_READ | pike.smb2.DELETE),
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        )
        nb_req = create_req1.parent.parent
        write_req = chan.write_request(RelatedOpen(tree), 0, "Expect fail")
        nb_req.adopt(write_req.parent)

        close_req = chan.close_request(RelatedOpen(tree))
        nb_req.adopt(close_req.parent)

        with self.assert_error(pike.ntstatus.STATUS_ACCESS_DENIED):
            (create_res1, write_res, close_res) = chan.connection.transceive(nb_req)

    def test_create_failed_and_query(self):
        chan, tree = self.tree_connect()
        name = "create_query_failed"
        test_dir = chan.create(
            tree,
            name,
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
        ).result()

        create_req = chan.create_request(
            tree,
            name,
            access=pike.smb2.GENERIC_READ,
            disposition=pike.smb2.FILE_CREATE,
            options=pike.smb2.FILE_DIRECTORY_FILE,
        )
        nb_req = create_req.parent.parent

        query_req = chan.query_directory_request(RelatedOpen(tree))
        nb_req.adopt(query_req.parent)

        (create_future, query_future) = chan.connection.submit(nb_req)
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_COLLISION):
            create_future.result()

        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_COLLISION):
            query_future.result()

        chan.close(test_dir)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#       reparse.py
#
# Abstract:
#
#       Set/Get reparse point tests
#
# Authors: Masen Furer (masen.furer@emc.com)
#

import pike.model
import pike.test

share_all = (
    pike.smb2.FILE_SHARE_READ | pike.smb2.FILE_SHARE_WRITE | pike.smb2.FILE_SHARE_DELETE
)


class TestReparsePoint(pike.test.PikeTest):
    def test_set_get_reparse_point(self):
        chan, tree = self.tree_connect()

        target = "target"

        link = chan.create(
            tree,
            "symlink",
            pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE,
            options=pike.smb2.FILE_OPEN_REPARSE_POINT | pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()
        chan.set_symlink(link, target, pike.smb2.SYMLINK_FLAG_RELATIVE)
        result = chan.get_symlink(link)
        self.assertEqual(result[0][0].reparse_data.substitute_name, target)
        chan.close(link)

    def test_symbolic_link_error_response(self):
        chan, tree = self.tree_connect()

        target = "target"

        link = chan.create(
            tree,
            "symlink",
            pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE,
            share=share_all,
            options=pike.smb2.FILE_OPEN_REPARSE_POINT | pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()
        chan.set_symlink(link, target, pike.smb2.SYMLINK_FLAG_RELATIVE)
        try:
            chan.create(tree, "symlink", share=share_all).result()
        except pike.model.ResponseError as err:
            self.assertEqual(
                err.response.status, pike.ntstatus.STATUS_STOPPED_ON_SYMLINK
            )
            self.assertEqual(err.response[0][0].error_data.substitute_name, target)
            self.assertEqual(err.response[0][0].error_data.print_name, target)
        chan.close(link)



#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        auth.py
#
# Abstract:
#
#        Authentication Plugins for Pike
#
# Authors: Masen Furer (masen.furer@dell.com)
#

"""
Authentication Plugins for Pike

This module contains wrappers around external authentication mechanisms and APIs.
"""

from __future__ import absolute_import

from builtins import object
import array
import warnings

try:
    from . import kerberos
except ImportError:
    kerberos = None
try:
    from . import ntlm
except ImportError:
    ntlm = None


def split_credentials(creds):
    if isinstance(creds, bytes):
        warnings.warn(
            "Pass creds as unicode string, got {!r}".format(creds), UnicodeWarning
        )
        creds = creds.decode("utf-8")
    user, password = creds.split("%", 1)
    if "\\" in user:
        domain, user = user.split("\\", 1)
    else:
        domain = "NONE"
    return (domain, user, password)


class KerberosProvider(object):
    def __init__(self, conn, creds=None):
        if creds:
            # XXX: NTLM support is only provided in likewise gssapi
            raise NotImplementedError("NTLM via GSSAPI is not functional")
            # This API doesn't accept strings with null terminators, so it cannot take
            # utf-16-le (which is what would be expected), so prefer instead to be
            # safe and only allow ascii characters.
            cred_encoding = "ascii"
            domain, user, password = split_credentials(creds)
            (self.result, self.context) = kerberos.authGSSClientInit(
                "cifs/" + conn.server,
                gssmech=2,
                user=user.encode(cred_encoding),
                password=password.encode(cred_encoding),
                domain=domain.encode(cred_encoding),
            )
        else:
            (self.result, self.context) = kerberos.authGSSClientInit(
                "cifs/" + conn.server, gssmech=1
            )

    def step(self, sec_buf):
        self.result = kerberos.authGSSClientStep(self.context, sec_buf.tobytes())
        if self.result == 0:
            return (
                array.array("B", kerberos.authGSSClientResponse(self.context)),
                None,
            )
        else:
            kerberos.authGSSClientSessionKey(self.context)
            return (
                None,
                array.array("B", kerberos.authGSSClientResponse(self.context)[:16]),
            )

    def username(self):
        return kerberos.authGSSClientUserName(self.context)


class NtlmProvider(object):
    def __init__(self, conn, creds):
        self.authenticator = ntlm.NtlmAuthenticator(*split_credentials(creds))

    def step(self, sec_buf):
        if self.authenticator.negotiate_message is None:
            return (self.authenticator.negotiate(), None)
        elif self.authenticator.challenge_message is None:
            self.authenticator.authenticate(sec_buf)
        return (
            self.authenticator.authenticate_buffer,
            self.authenticator.exported_session_key,
        )

    def username(self):
        return "{0}\{1}".format(self.authenticator.domain, self.authenticator.username)



#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        crypto.py
#
# Abstract:
#
#        SMB3 Encryption
#
# Authors: Masen Furer (masen.furer@dell.com)
#

from __future__ import absolute_import
from builtins import map
from builtins import range
from builtins import object

import array
import random

try:
    from Crypto.Cipher import AES
except ImportError:
    from Cryptodome.Cipher import AES

from . import core
from . import digest
from . import smb2


def random_bytes(length):
    """
    Generate an array of random bytes.

    :param length: How many bytes to return
    :return: array.array('B') of length random bytes
    """
    return array.array("B", (random.randint(0, 255) for _ in range(length)))


def pad_right(value, length, byte=b"\0"):
    if len(value) > length:
        value = value[:length]
    elif len(value) < 16:
        value += array.array("B", byte * (length - len(value)))
    return value


class CipherMismatch(Exception):
    pass


class Ciphers(core.ValueEnum):
    SMB2_NONE_CIPHER = 0x0000
    SMB2_AES_128_CCM = 0x0001
    SMB2_AES_128_GCM = 0x0002


Ciphers.import_items(globals())

cipher_map = {
    SMB2_AES_128_CCM: (AES.MODE_CCM, 11),
    SMB2_AES_128_GCM: (AES.MODE_GCM, 12),
}


class EncryptionCapabilities(core.Frame):
    context_type = smb2.SMB2_ENCRYPTION_CAPABILITIES

    def __init__(self):
        self.ciphers = []
        self.ciphers_count = None

    def _encode(self, cur):
        if self.ciphers_count is None:
            self.ciphers_count = len(self.ciphers)
        cur.encode_uint16le(self.ciphers_count)
        for c in self.ciphers:
            cur.encode_uint16le(c)

    def _decode(self, cur):
        self.ciphers_count = cur.decode_uint16le()
        for ix in range(self.ciphers_count):
            self.ciphers.append(Ciphers(cur.decode_uint16le()))


class EncryptionCapabilitiesRequest(
    smb2.NegotiateRequestContext, EncryptionCapabilities
):
    def __init__(self, parent):
        smb2.NegotiateRequestContext.__init__(self, parent)
        EncryptionCapabilities.__init__(self)


class EncryptionCapabilitiesResponse(
    smb2.NegotiateResponseContext, EncryptionCapabilities
):
    def __init__(self, parent):
        smb2.NegotiateResponseContext.__init__(self, parent)
        EncryptionCapabilities.__init__(self)


class TransformHeader(core.Frame):
    """
    TransformHeader is designed to be a transparent slip attached to a Netbios
    frame object (specified as parent). During serialization, if a Netbios frame
    contains an attribute "transform", then that object's encode method will be
    called instead of each child frames' encode method. This frame is responsible
    for both encrypting and decrypting Smb2 frames according to an attached
    encryption context.

    If the encryption_context is not explicitly specified, then it will be looked
    up based on session_id from the parent Netbios frame's connection reference
    """

    LOG_CHILDREN_COUNT = False
    LOG_CHILDREN_EXPAND = True

    def __init__(self, parent):
        core.Frame.__init__(self, parent)
        self.protocol_id = array.array("B", b"\xfdSMB")
        self.signature = None
        # the value of nonce is always used in the encryption routine
        self.nonce = array.array("B", map(random.randint, [0] * 16, [255] * 16))
        # if wire_nonce is set, it will be sent on the wire instead of nonce
        self.wire_nonce = None
        self.original_message_size = None
        self.reserved = None
        self.flags = 0x1
        self.session_id = None
        self.encryption_context = None
        self.additional_authenticated_data_buf = array.array("B")
        if parent is not None:
            parent.transform = self
        else:
            self._smb2_frames = []

    def _log_str(self):
        components = [type(self).__name__]
        if self.children:
            components.extend(self._log_str_children())
        return " ".join(components)

    def _children(self):
        if self.parent is not None:
            return self.parent._children()
        else:
            return self._smb2_frames

    def append(self, smb2_frame):
        if self.parent is not None:
            self.parent.append(smb2_frame)
        else:
            self._smb2_frames.append(smb2_frame)

    def _encode(self, cur):
        self._encode_header(cur)
        self._encode_smb2(cur)

    def _encode_header(self, cur):
        # look up the encryption context if not specified
        if self.encryption_context is None and self.parent is not None:
            self.encryption_context = self.parent.conn.encryption_context(
                self.session_id
            )
        cur.encode_bytes(self.protocol_id)

        # the signature will be written in _encode_smb2
        self.signature_offset = cur.offset
        cur.advanceto(cur + 16)
        # the crypto header will be written in _encode_smb2
        self.crypto_header_offset = cur.offset
        # save a space for the wire nonce
        self.wire_nonce_hole = cur.hole.encode_bytes(b"\0" * 16)
        cur.advanceto(cur + 16)

        # the following fields are part of AdditionalAuthenticatedData and are
        # used as inputs to the AES cipher
        aad_cur = core.Cursor(self.additional_authenticated_data_buf, 0)
        # nonce field size is 16 bytes right padded with zeros
        self.nonce = pad_right(self.nonce, 16)
        aad_cur.encode_bytes(self.nonce)
        self.original_message_size_hole = aad_cur.hole.encode_uint32le(0)
        if self.reserved is None:
            self.reserved = 0
        aad_cur.encode_uint16le(self.reserved)  # reserved
        aad_cur.encode_uint16le(self.flags)
        aad_cur.encode_uint64le(self.session_id)

    def _encode_smb2(self, cur):
        # serialize all chained Smb2 commands into one buffer
        original_message_buf = array.array("B")
        original_message_cur = core.Cursor(original_message_buf, 0)
        for smb_frame in self.parent:
            smb_frame.encode(original_message_cur)
        if self.original_message_size is None:
            self.original_message_size = len(original_message_buf)
        self.original_message_size_hole(self.original_message_size)
        (self.ciphertext, crypto_hmac) = self.encryption_context.encrypt(
            original_message_buf, self.additional_authenticated_data_buf, self.nonce
        )
        cur.encode_bytes(self.ciphertext)

        # fill in the signature hole
        sig_cur = core.Cursor(cur.array, self.signature_offset)
        if self.signature is None:
            self.signature = crypto_hmac
        sig_cur.encode_bytes(self.signature)

        # fill in the header
        aad_cur = core.Cursor(cur.array, self.crypto_header_offset)
        aad_cur.encode_bytes(self.additional_authenticated_data_buf)

        # fill in the wire nonce
        if self.wire_nonce is None:
            self.wire_nonce = self.nonce
        self.wire_nonce_hole(pad_right(self.wire_nonce, 16))

    def _decode(self, cur):
        self._decode_header(cur)
        self._decode_smb2(cur)

    def _decode_header(self, cur):
        self.protocol_id = cur.decode_bytes(4)
        self.signature = cur.decode_bytes(16)
        # the following fields are part of AdditionalAuthenticatedData and are
        # used as inputs to the AES cipher
        self.crypto_header_start = cur.offset
        self.nonce = cur.decode_bytes(16)
        self.original_message_size = cur.decode_uint32le()
        self.reserved = cur.decode_uint16le()  # reserved
        self.flags = cur.decode_uint16le()
        self.session_id = cur.decode_uint64le()
        self.crypto_header_end = cur.offset
        if self.encryption_context is None and self.parent is not None:
            self.encryption_context = self.parent.conn.encryption_context(
                self.session_id
            )

    def _decode_smb2(self, cur):
        self.encrypted_data = cur.decode_bytes(self.original_message_size)
        crypto_header = cur.array[self.crypto_header_start : self.crypto_header_end]
        self.plaintext = self.encryption_context.decrypt(
            self.encrypted_data, self.signature, crypto_header, self.nonce
        )
        # scan through the plaintext for chained messages
        pt_cur = core.Cursor(self.plaintext, 0)
        end = pt_cur + len(self.plaintext)
        with pt_cur.bounded(pt_cur, end):
            while pt_cur < end:
                start = pt_cur.offset
                message = smb2.Smb2(self.parent)
                message.decode(pt_cur)
                message.buf = pt_cur.array[start : pt_cur.offset]

    def verify(self, *args, **kwds):
        pass  # verification occurs at the point of decryption


class CryptoKeys300(object):
    """Key generation for SMB 0x300 and 0x302"""

    def __init__(self, session_key, *args, **kwds):
        self.encryption = digest.derive_key(session_key, b"SMB2AESCCM", b"ServerIn \0")[
            :16
        ].tobytes()
        self.decryption = digest.derive_key(session_key, b"SMB2AESCCM", b"ServerOut\0")[
            :16
        ].tobytes()


class CryptoKeys311(object):
    """Key generation for SMB 0x311 +"""

    def __init__(self, session_key, pre_auth_integrity_hash, *args, **kwds):
        self.encryption = digest.derive_key(
            session_key, b"SMBC2SCipherKey", pre_auth_integrity_hash
        )[:16].tobytes()
        self.decryption = digest.derive_key(
            session_key, b"SMBS2CCipherKey", pre_auth_integrity_hash
        )[:16].tobytes()


class EncryptionContext(object):
    """
    Encapsulates all information needed to encrypt and decrypt messages.
    This context is attached to an SMB Session object
    """

    def __init__(self, keys, ciphers):
        self.keys = keys
        for c in ciphers:
            if c in cipher_map:
                self.aes_mode, self.nonce_length = cipher_map[c]
                self.cipher = c
                break
        else:
            raise CipherMismatch(
                "Client did not recognize any ciphers returned by the "
                "server: {0}".format(ciphers)
            )

    def encrypt(self, plaintext, authenticated_data, nonce):
        enc_cipher = AES.new(
            self.keys.encryption,
            self.aes_mode,
            nonce=nonce[: self.nonce_length].tobytes(),
        )
        enc_cipher.update(authenticated_data.tobytes())
        ciphertext, signature = enc_cipher.encrypt_and_digest(plaintext.tobytes())
        return array.array("B", ciphertext), array.array("B", signature)

    def decrypt(self, ciphertext, signature, authenticated_data, nonce):
        dec_cipher = AES.new(
            self.keys.decryption,
            self.aes_mode,
            nonce=nonce[: self.nonce_length].tobytes(),
        )
        dec_cipher.update(authenticated_data.tobytes())
        return array.array(
            "B",
            dec_cipher.decrypt_and_verify(ciphertext.tobytes(), signature.tobytes()),
        )



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        multichannel.py
#
# Abstract:
#
#        Multichannel tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.model
import pike.smb2
import pike.test
import pike.ntstatus


@pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
@pike.test.RequireCapabilities(pike.smb2.SMB2_GLOBAL_CAP_MULTI_CHANNEL)
class MultiChannelTest(pike.test.PikeTest):
    def session_bind(self, chan):
        return (
            chan.connection.client.connect(self.server)
            .negotiate()
            .session_setup(self.creds, bind=chan.session)
        )

    # Open a file, binding session to second channel, and close it there
    def test_open_bind_close(self):
        chan, tree = self.tree_connect()
        handle = chan.create(tree, "hello.txt").result()

        # Open a second channel and bind the session
        chan2 = self.session_bind(chan)

        # Close the handle on the second channel
        chan2.close(handle)

    # Simulate a channel failover during a write and confirm that a
    # stale write sent on the original channel is rejected
    def test_write_fence_reject_stale(self):
        data_stale = b"stale"
        data_fresh = b"fresh"
        chan, tree = self.tree_connect()
        client = chan.connection.client

        # Open a file
        handle = chan.create(
            tree, "hello.txt", disposition=pike.smb2.FILE_SUPERSEDE
        ).result()

        # Open a second channel
        chan2 = self.session_bind(chan)

        # Pretend channel failover occured
        client.channel_sequence = 1

        # Send write replay
        with chan2.let(flags=pike.smb2.SMB2_FLAGS_REPLAY_OPERATION):
            chan2.write(handle, 0, data_stale)

        # Send normal write
        chan2.write(handle, 0, data_fresh)

        # Send stale write on original channel, which should fail.
        # Spec is ambiguous as to status code, but Windows Server 2012
        # seems to return STATUS_FILE_NOT_AVAILABLE
        with self.assert_error(pike.ntstatus.STATUS_FILE_NOT_AVAILABLE):
            with chan.let(channel_sequence=0):
                chan.write(handle, 0, data_stale)

        # Read back data to ensure it is not stale value
        result = chan2.read(handle, len(data_fresh), 0).tobytes()
        self.assertEquals(result, data_fresh)

        # Close handle
        chan2.close(handle)

        # Close second connection
        chan2.connection.close()



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        model.py
#
# Abstract:
#
#        Transport and object model
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#          Masen Furer (masen.furer@dell.com)
#

"""
SMB2 Object Model.

This module contains an implementation of the SMB2 client object model,
allowing channels, sessions, tree connections, opens, and leases
to be established and tracked.  It provides convenience functions
for exercising common elements of the protocol without manually
constructing packets.
"""

from __future__ import absolute_import
from builtins import next
from builtins import map
from builtins import range
from builtins import object
from builtins import str
from future.utils import raise_
from future.utils import raise_from

import array
import contextlib
from functools import reduce
import logging
import operator
import random
import socket
import struct
import sys
import time
import warnings

from . import auth
from . import core
from . import crypto
from . import netbios
from . import nttime
from . import smb2
from . import transport
from . import ntstatus
from . import digest
from .exceptions import (
    CallbackError,
    CreditError,
    RequestError,
    ResponseError,
    StateError,
    TimeoutError,
)
import pike.io

default_credit_request = 10
default_timeout = 30
default_port = 445
trace = False
Open = pike.io.CompatOpen  # maintain the old __init__ signature of pike.model.Open


def loop(timeout=None, count=None):
    """
    wrapper for blocking on the underlying event loop for the given timeout
    or given count of iterations
    """
    if timeout is None:
        timeout = default_timeout
    transport.loop(timeout=timeout, count=count)


class AssertNtstatusContext(object):
    response = None
    exception = None


@contextlib.contextmanager
def pike_status(exp_status):
    obj = AssertNtstatusContext()
    try:
        yield obj
        if exp_status != ntstatus.STATUS_SUCCESS:
            raise AssertionError(
                "No error raised when " "ResponseError({}) expected.".format(exp_status)
            )
    except ResponseError as err:
        obj.exception = err
        obj.response = err.response
        if err.response.status != exp_status:
            raise_from(
                AssertionError(
                    "{} raised when "
                    "expecting ResponseError({})".format(
                        err.response.status, exp_status
                    )
                ),
                err,
            )


class Events(core.ValueEnum):
    """Events used for callback functions"""

    EV_REQ_PRE_SERIALIZE = 0x1  # cb expects Netbios frame
    EV_REQ_POST_SERIALIZE = 0x2  # cb expects Netbios frame
    EV_REQ_PRE_SEND = 0x3  # cb expects a buffer to send
    EV_REQ_POST_SEND = 0x4  # cb expects an integer of bytes sent
    EV_RES_PRE_RECV = 0x5  # cb expects an integer of bytes to read
    EV_RES_POST_RECV = 0x6  # cb expects a buffer that was read
    EV_RES_PRE_DESERIALIZE = 0x7  # cb expects a complete netbios buffer
    EV_RES_POST_DESERIALIZE = 0x8  # cb expects a Netbios frame


Events.import_items(globals())


class Future(object):
    """
    Result of an asynchronous operation.

    Futures represent the result of an operation that has not yet completed.
    In Pike, they are most commonly used to track SMB2 request/response pairs,
    but they can be used for any asynchronous operation.

    The result of a future can be waited for synchronously by simply calling
    L{Future.result}, or a notification callback can be set with L{Future.then}.

    Futures implement the context manager interface so that they can be used
    as the context for a with block.  If an exception is raised from the block,
    it will automatically be set as the result of the future.

    @ivar request: The request associated with the future, usually an SMB2 request frame.
    @ivar response: The result of the future, usually an SMB2 response frame.
    @ivar interim_response: The interim response, usually an SMB2 response frame.
    @ivar traceback: The traceback of an exception result, if applicable.
    """

    def __init__(self, request=None):
        """
        Initialize future.

        @param request: The request associated with the response.
        """
        self.request = request
        self.interim_response = None
        self.response = None
        self.notify = None
        self.traceback = None

    def complete(self, response, traceback=None):
        """
        Completes the future with the given result.

        Calling a future as a function is equivalent to calling this method.

        @param response: The result of the future.
        @param traceback: If response is an exception, an optional traceback
        """
        self.response = response
        self.traceback = traceback
        if self.notify is not None:
            self.notify(self)

    def interim(self, response):
        """
        Set interim response.

        @param response: The interim response.
        """
        self.interim_response = response

    def wait(self, timeout=None):
        """
        Wait for future result to become available.

        @param timeout: The time in seconds before giving up and raising TimeoutError
        """
        if timeout is None:
            timeout = default_timeout
        deadline = time.time() + timeout
        while self.response is None:
            now = time.time()
            if now > deadline:
                raise TimeoutError.with_future(
                    self, "Timed out after %s seconds" % timeout
                )
            loop(timeout=deadline - now, count=1)

        return self

    def wait_interim(self, timeout=None):
        """
        Wait for interim response or actual result to become available.

        @param timeout: The time in seconds before giving up and raising TimeoutError
        """
        if timeout is None:
            timeout = default_timeout
        deadline = time.time() + timeout
        while self.response is None and self.interim_response is None:
            now = time.time()
            if now > deadline:
                raise TimeoutError.with_future(
                    self, "Timed out after %s seconds" % timeout
                )
            loop(timeout=deadline - now, count=1)

        return self

    def result(self, timeout=None):
        """
        Return result of future.

        If the result is not yet available, this function will wait for it.
        If the result is an exception, this function will raise it instead of
        returning it.

        @param timeout: The time in seconds before giving up and raising TimeoutError
        """
        self.wait(timeout)

        if isinstance(self.response, BaseException):
            traceback = self.traceback
            self.traceback = None
            raise_(self.response, None, traceback)
        else:
            return self.response

    def then(self, notify):
        """
        Set notification function.

        @param notify: A function which will be invoked with this future as a parameter
                       when its result becomes available.  If it is already available,
                       it will be called immediately.
        """
        if not callable(notify):
            raise CallbackError("{0} is not a callable object".format(notify))
        if self.response is not None:
            notify(self)
        else:
            self.notify = notify

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            self.complete(exc_value, traceback)
            return True

    def __call__(self, *params, **kwparams):
        self.complete(*params, **kwparams)


class Client(object):
    """
    Maintains all state associated with an SMB2/3 client.

    :type dialects: Sequence[pike.smb2.Dialect]
    :ivar dialects: A list of supported dialects
    :ivar capabilities: Capabilities flags
    :ivar security_mode: Security mode flags
    :ivar client_guid: Client GUID
    :ivar channel_sequence: Current channel sequence number
    """

    def __init__(
        self,
        dialects=[
            smb2.DIALECT_SMB2_002,
            smb2.DIALECT_SMB2_1,
            smb2.DIALECT_SMB3_0,
            smb2.DIALECT_SMB3_0_2,
            smb2.DIALECT_SMB3_1_1,
        ],
        capabilities=smb2.GlobalCaps(reduce(operator.or_, smb2.GlobalCaps.values())),
        security_mode=smb2.SMB2_NEGOTIATE_SIGNING_ENABLED,
        client_guid=None,
    ):
        if client_guid is None:
            client_guid = array.array("B", map(random.randint, [0] * 16, [255] * 16))

        self.dialects = dialects
        self.capabilities = capabilities
        self.security_mode = security_mode
        self.client_guid = client_guid
        self.channel_sequence = 0
        self.callbacks = {}
        self._oplock_break_map = {}
        self._lease_break_map = {}
        self._oplock_break_queue = []
        self._lease_break_queue = []
        self._connections = []
        self._leases = {}

        self.logger = logging.getLogger("pike")

    @contextlib.contextmanager
    def callback(self, event, cb):
        """
        Register a callback function for the context block, then unregister it
        """
        self.register_callback(event, cb)
        try:
            yield
        finally:
            self.unregister_callback(event, cb)

    def register_callback(self, event, cb):
        """
        Registers a callback function, cb for the given event.
        When the event fires, cb will be called with the relevant top-level
        Netbios frame as the single paramter.
        """
        ev = Events(event)
        if ev not in self.callbacks:
            self.callbacks[ev] = []
        self.callbacks[ev].append(cb)

    def unregister_callback(self, event, cb):
        """
        Unregisters a callback function, cb for the given event.
        """
        ev = Events(event)
        if ev not in self.callbacks:
            return
        if cb not in self.callbacks[ev]:
            return
        self.callbacks[ev].remove(cb)

    def restrict_dialects(self, min_dialect=None, max_dialect=None):
        """
        Update the list of SMB2 dialects this Client will advertise support for.

        This call will only remove dialects from the list.

        Dialects should be from pike.smb2.Dialect enum, but technically may be any
        int or float.

        If either of min_dialect or max_dialect is None, the minimum or maximum dialect
        supported will be used.

        @param min_dialect: The minimum dialect to support (inclusive)
        @param max_dialect: The maximum dialect to support (inclusive)
        """
        if min_dialect is None:
            min_dialect = min(smb2.Dialect.values())
        if max_dialect is None:
            max_dialect = max(smb2.Dialect.values())
        self.dialects = [d for d in self.dialects if min_dialect <= d <= max_dialect]

    def connect(self, server, port=default_port):
        """
        Create a connection.

        Returns a new L{Connection} object connected to the given server and port.

        @param server: The server to connect to.
        @param port: The port to connect to.
        """
        return self.connect_submit(server, port).result()

    def connect_submit(self, server, port=default_port):
        """
        Create a connection.

        Returns a new L{Future} object for the L{Connection} being established
        asynchronously to the given server and port.

        @param server: The server to connect to.
        @param port: The port to connect to.
        """
        return Connection(self, server, port).connection_future

    # Do not use, may be removed.  Use oplock_break_future.
    def next_oplock_break(self):
        while len(self._oplock_break_queue) == 0:
            loop(count=1)
        return self._oplock_break_queue.pop()

    # Do not use, may be removed.  Use lease_break_future.
    def next_lease_break(self):
        while len(self._lease_break_queue) == 0:
            loop(count=1)
        return self._lease_break_queue.pop()

    def oplock_break_future(self, file_id):
        """
        Create future for oplock break.

        Returns a L{Future} object which will be completed when
        an oplock break occurs.  The result will be the L{smb2.Smb2} frame
        of the break notification packet.

        @type file_id: (number, number)
        @param file_id: The file ID of the oplocked file.
        """

        future = Future(request=("OplockBreak", file_id))

        for smb_res in self._oplock_break_queue[:]:
            if smb_res[0].file_id == file_id:
                future.complete(smb_res)
                self._oplock_break_queue.remove(smb_res)
                break

        if future.response is None:
            self._oplock_break_map[file_id] = future

        return future

    def lease_break_future(self, lease_key):
        """
        Create future for lease break.

        Returns a L{Future} object which will be completed when
        a lease break occurs.  The result will be the L{smb2.Smb2} frame
        of the break notification packet.

        @param lease_key: The lease key for the lease.
        """

        future = Future(request=("LeaseBreak", core.Frame._value_str(lease_key)))

        for smb_res in self._lease_break_queue[:]:
            if smb_res[0].lease_key == lease_key:
                future.complete(smb_res)
                self._lease_break_queue.remove(smb_res)
                break

        if future.response is None:
            self._lease_break_map[lease_key.tobytes()] = future

        return future

    def oplock_break(self, file_id):
        """
        Wait for and return oplock break notification.

        Equivalent to ``oplock_break_future(file_id).result()``
        """

        return self.oplock_break_future(file_id).result()

    def lease_break(self, lease_key):
        """
        Wait for and return lease break notification.

        Equivalent to ``lease_break_future(file_id).result()``
        """

        return self.lease_break_future(lease_key).result()

    def lease(self, tree, lease_res):
        """
        Create or look up lease object.

        Returns a lease object based on a L{Tree} and a
        L{smb2.LeaseResponse}.  The lease object is created
        if it does not already exist.

        @param tree: The tree on which the lease request was issued.
        @param lease_res: The lease create context response.
        """

        lease_key = lease_res.lease_key.tobytes()
        if lease_key not in self._leases:
            lease = Lease(tree)
            self._leases[lease_key] = lease
        else:
            lease = self._leases[lease_key]
            lease.ref()

        lease.update(lease_res)
        return lease

    # Internal function to remove lease from table
    def dispose_lease(self, lease):
        del self._leases[lease.lease_key.tobytes()]


class Connection(transport.Transport):
    """
    Connection to server.

    Represents a connection to a server and handles all socket operations
    and request/response dispatch.

    :type client: Client
    :ivar client: The Client object associated with this connection.
    :ivar server: The server name or address
    :ivar port: The server port
    """

    def __init__(self, client, server, port=default_port):
        """
        Constructor.

        This should generally not be used directly.  Instead,
        use L{Client.connect}().
        """
        super(Connection, self).__init__()
        self._no_delay = True
        self._in_buffer = array.array("B")
        self._watermark = 4
        self._out_buffer = None
        self._next_mid = 0
        self._mid_blacklist = set()
        self._out_queue = []
        self._future_map = {}
        self._sessions = {}
        self._binding = None
        self._binding_key = None
        self._settings = {}
        self._pre_auth_integrity_hash = array.array("B", b"\0" * 64)
        self._negotiate_request = None
        self._negotiate_response = None
        self.callbacks = {}
        self.connection_future = Future(request=(server, port))
        self.credits = 0
        self.client = client
        self.server = server
        self.port = port
        self.remote_addr = None
        self.local_addr = None
        self.verify_signature = True

        self.error = None
        self.traceback = None

        for result in socket.getaddrinfo(
            server, port, 0, socket.SOCK_STREAM, socket.IPPROTO_TCP
        ):
            family, socktype, proto, canonname, sockaddr = result
            break
        self.create_socket(family, socktype)
        self.connect(sockaddr)

    @contextlib.contextmanager
    def callback(self, event, cb):
        """
        Register a callback function for the context block, then unregister it
        """
        self.register_callback(event, cb)
        try:
            yield
        finally:
            self.unregister_callback(event, cb)

    def register_callback(self, event, cb):
        """
        Registers a callback function, cb for the given event.
        When the event fires, cb will be called with the relevant top-level
        Netbios frame as the single paramter.
        """
        ev = Events(event)
        if ev not in self.callbacks:
            self.callbacks[ev] = []
        self.callbacks[ev].append(cb)

    def unregister_callback(self, event, cb):
        """
        Unregisters a callback function, cb for the given event.
        """
        ev = Events(event)
        if ev not in self.callbacks:
            return
        if cb not in self.callbacks[ev]:
            return
        self.callbacks[ev].remove(cb)

    def process_callbacks(self, event, obj):
        """
        Fire callbacks for the given event, passing obj as the parameter

        Connection-specific callbacks will be fired first, followed by client
        callbacks
        """
        ev = Events(event)
        all_callbacks = [self.callbacks]
        if hasattr(self.client, "callbacks"):
            all_callbacks.append(self.client.callbacks)
        for callbacks in all_callbacks:
            if ev not in callbacks:
                continue
            for cb in callbacks[ev]:
                cb(obj)

    @property
    def negotiate_response(self):
        return self._negotiate_response

    @negotiate_response.setter
    def negotiate_response(self, value):
        if self._negotiate_response is not None:
            raise AssertionError("Attempting to overwrite negotiate response")
        self._negotiate_response = value
        # pre-auth integrity hash processing
        if value.dialect_revision < smb2.DIALECT_SMB3_1_1:
            return
        self._pre_auth_integrity_hash = digest.smb3_sha512(
            self._pre_auth_integrity_hash + self._negotiate_request.parent.serialize()
        )
        self._pre_auth_integrity_hash = digest.smb3_sha512(
            self._pre_auth_integrity_hash + value.parent.parent.buf[4:]
        )

    @property
    def dialect_revision(self):
        return (
            self.negotiate_response.dialect_revision
            if self.negotiate_response is not None
            else 0x0
        )

    @property
    def hostname(self):
        port = ""
        if self.port != default_port:
            port = ":{}".format(self.port)
        return self.server + port

    def next_mid_range(self, length):
        """
        multicredit requests must reserve 1 message id per credit charged.
        the message id of the request should be the first id of the range.
        """
        if length < 1:
            length = 1
        start_range = self._next_mid
        while True:
            r = set(range(start_range, start_range + length))
            if not r.intersection(self._mid_blacklist):
                break
            start_range += 1
        self._next_mid = sorted(list(r))[-1] + 1
        return start_range

    def next_mid(self):
        return self.next_range(1)

    def reserve_mid(self, mid):
        self._mid_blacklist.add(mid)

    def handle_connect(self):
        self.client._connections.append(self)
        with self.connection_future:
            self.local_addr = self.socket.getsockname()
            self.remote_addr = self.socket.getpeername()

            self.client.logger.debug(
                "connect: %s/%s -> %s/%s",
                self.local_addr[0],
                self.local_addr[1],
                self.remote_addr[0],
                self.remote_addr[1],
            )
        self.connection_future(self)

    def handle_read(self):
        # Try to read the next netbios frame
        remaining = self._watermark - len(self._in_buffer)
        self.process_callbacks(EV_RES_PRE_RECV, remaining)
        data = array.array("B", self.recv(remaining))
        self.process_callbacks(EV_RES_POST_RECV, data)
        self._in_buffer.extend(data)
        avail = len(self._in_buffer)
        if avail >= 4:
            self._watermark = 4 + struct.unpack(">L", self._in_buffer[0:4])[0]
        if avail == self._watermark:
            nb = self.frame()
            self.process_callbacks(EV_RES_PRE_DESERIALIZE, self._in_buffer)
            nb.parse(self._in_buffer)
            self._in_buffer = array.array("B")
            self._watermark = 4
            self._dispatch_incoming(nb)

    def handle_write(self):
        # Try to write out more data
        while self._out_buffer is None and len(self._out_queue):
            self._out_buffer = self._prepare_outgoing()
            while self._out_buffer is not None:
                self.process_callbacks(EV_REQ_PRE_SEND, self._out_buffer)
                sent = self.send(self._out_buffer)
                del self._out_buffer[:sent]
                if len(self._out_buffer) == 0:
                    self._out_buffer = None
                self.process_callbacks(EV_REQ_POST_SEND, sent)

    def handle_close(self):
        self.close()

    def handle_error(self):
        (_, self.error, self.traceback) = sys.exc_info()
        self.close()

    def close(self):
        """
        Close connection.

        This unceremoniously terminates the connection and fails all
        outstanding requests with EOFError.
        """
        # If there is no error, propagate EOFError
        if self.error is None:
            self.error = EOFError("close")

        # if the connection hasn't been established, raise the error
        if self.connection_future.response is None:
            self.connection_future(self.error)

        # otherwise, ignore this connection since it's not associated with its client
        if self not in self.client._connections:
            return

        super(Connection, self).close()

        if self.remote_addr is not None:
            self.client.logger.debug(
                "disconnect (%s/%s -> %s/%s): %s",
                self.local_addr[0],
                self.local_addr[1],
                self.remote_addr[0],
                self.remote_addr[1],
                self.error,
            )

        self.client._connections.remove(self)

        for future in self._out_queue:
            future.complete(self.error, self.traceback)
        del self._out_queue[:]

        for future in self._future_map.values():
            future.complete(self.error, self.traceback)
        self._future_map.clear()

        for session in list(self._sessions.values()):
            session.delchannel(self)

        self.traceback = None

    def _prepare_outgoing(self):
        # Try to prepare an outgoing packet

        # Grab an outgoing smb2 request
        future = self._out_queue[0]

        result = None
        with future:
            req = future.request
            self.process_callbacks(EV_REQ_PRE_SERIALIZE, req.parent)

            if req.credit_charge is None:
                req.credit_charge = 0
                for cmd in req:
                    if isinstance(cmd, smb2.ReadRequest) and cmd.length > 0:
                        # special handling, 1 credit per 64k
                        req.credit_charge, remainder = divmod(
                            cmd.length, smb2.BYTES_PER_CREDIT
                        )
                    elif isinstance(cmd, smb2.WriteRequest) and cmd.buffer is not None:
                        # special handling, 1 credit per 64k
                        if cmd.length is None:
                            cmd.length = len(cmd.buffer)
                        req.credit_charge, remainder = divmod(
                            cmd.length, smb2.BYTES_PER_CREDIT
                        )
                    else:
                        remainder = 1  # assume 1 credit per command
                    if remainder > 0:
                        req.credit_charge += 1
            # do credit accounting based on our calculations (MS-SMB2 3.2.5.1)
            self.credits -= req.credit_charge

            if req.credit_request is None:
                req.credit_request = default_credit_request
                if req.credit_charge > req.credit_request:
                    req.credit_request = req.credit_charge  # try not to fall behind

            del self._out_queue[0]

            # Assign message id
            if req.message_id is None:
                req.message_id = self.next_mid_range(req.credit_charge)

            if req.is_last_child():
                # Last command in chain, ready to send packet
                result = req.parent.serialize()
                self.process_callbacks(EV_REQ_POST_SERIALIZE, req.parent)
                if trace:
                    self.client.logger.debug(
                        "send (%s/%s -> %s/%s): %s",
                        self.local_addr[0],
                        self.local_addr[1],
                        self.remote_addr[0],
                        self.remote_addr[1],
                        req.parent,
                    )
                else:
                    self.client.logger.debug(
                        "send (%s/%s -> %s/%s): %s",
                        self.local_addr[0],
                        self.local_addr[1],
                        self.remote_addr[0],
                        self.remote_addr[1],
                        req.parent._log_str(),
                    )
            else:
                # Not ready to send chain
                result = None

            # Move it to map for response waiters (but not cancel)
            if not isinstance(req[0], smb2.Cancel):
                self._future_map[req.message_id] = future

        return result

    def _find_oplock_future(self, file_id):
        if file_id in self.client._oplock_break_map:
            return self.client._oplock_break_map.pop(file_id)
        return None

    def _find_lease_future(self, lease_key):
        lease_key = lease_key.tobytes()
        if lease_key in self.client._lease_break_map:
            return self.client._lease_break_map.pop(lease_key)
        return None

    def _dispatch_incoming(self, res):
        if trace:
            self.client.logger.debug(
                "recv (%s/%s -> %s/%s): %s",
                self.remote_addr[0],
                self.remote_addr[1],
                self.local_addr[0],
                self.local_addr[1],
                res,
            )
        else:
            self.client.logger.debug(
                "recv (%s/%s -> %s/%s): %s",
                self.remote_addr[0],
                self.remote_addr[1],
                self.local_addr[0],
                self.local_addr[1],
                res._log_str(),
            )
        self.process_callbacks(EV_RES_POST_DESERIALIZE, res)
        for smb_res in res:
            # TODO: move credit tracking to callbacks
            self.credits += smb_res.credit_response

            # Verify non-session-setup-response signatures
            # session setup responses are verified in SessionSetupContext
            if not isinstance(smb_res[0], smb2.SessionSetupResponse):
                key = self.signing_key(smb_res.session_id)
                if key and self.verify_signature:
                    smb_res.verify(self.signing_digest(), key)

            if smb_res.message_id == smb2.UNSOLICITED_MESSAGE_ID:
                if isinstance(smb_res[0], smb2.OplockBreakNotification):
                    future = self._find_oplock_future(smb_res[0].file_id)
                    if future:
                        future.complete(smb_res)
                    else:
                        self.client._oplock_break_queue.append(smb_res)
                elif isinstance(smb_res[0], smb2.LeaseBreakNotification):
                    future = self._find_lease_future(smb_res[0].lease_key)
                    if future:
                        future.complete(smb_res)
                    else:
                        self.client._lease_break_queue.append(smb_res)
                else:
                    raise core.BadPacket()
            elif smb_res.message_id in self._future_map:
                future = self._future_map[smb_res.message_id]
                if smb_res.status == ntstatus.STATUS_PENDING:
                    future.interim(smb_res)
                elif (
                    isinstance(smb_res[0], smb2.ErrorResponse)
                    or smb_res.status not in smb_res[0].allowed_status
                ):
                    future.complete(ResponseError(smb_res))
                    del self._future_map[smb_res.message_id]
                else:
                    future.complete(smb_res)
                    del self._future_map[smb_res.message_id]

    def submit(self, req):
        """
        Submit request.

        Submits a L{netbios.Netbios} frame for sending.  Returns
        a list of L{Future} objects, one for each corresponding
        L{smb2.Smb2} frame in the request.
        """
        if not isinstance(req, netbios.Netbios):
            raise RequestError(
                req, "{0} is not a netbios.Netbios frame".format(repr(req))
            )
        if self.error is not None:
            raise_(self.error, None, self.traceback)
        futures = []
        for smb_req in req:
            if isinstance(smb_req[0], smb2.Cancel):
                # Find original future being canceled to return
                if smb_req.async_id is not None:
                    # Cancel by async ID
                    future = [
                        f
                        for f in iter(self._future_map.values())
                        if f.interim_response.async_id == smb_req.async_id
                    ][0]
                elif smb_req.message_id in self._future_map:
                    # Cancel by message id, already in future map
                    future = self._future_map[smb_req.message_id]
                else:
                    # Cancel by message id, still in send queue
                    future = [
                        f
                        for f in self._out_queue
                        if f.request.message_id == smb_req.message_id
                    ][0]
                # Add fake future for cancel since cancel has no response
                self._out_queue.append(Future(request=smb_req))
                futures.append(future)
            else:
                future = Future(request=smb_req)
                self._out_queue.append(future)
                futures.append(future)

        # don't wait for the callback, send the data now
        if self._no_delay:
            self.handle_write()
        return futures

    def transceive(self, req):
        """
        Submit request and wait for responses.

        Submits a L{netbios.Netbios} frame for sending.  Waits for
        and returns a list of L{smb2.Smb2} response objects, one for each
        corresponding L{smb2.Smb2} frame in the request.
        """
        return [f.result() for f in self.submit(req)]

    def negotiate_request(self, hash_algorithms=None, salt=None, ciphers=None):
        smb_req = self.request()
        smb_req.credit_charge = 0  # negotiate requests are free
        neg_req = smb2.NegotiateRequest(smb_req)

        neg_req.dialects = self.client.dialects
        neg_req.security_mode = self.client.security_mode
        neg_req.capabilities = self.client.capabilities
        neg_req.client_guid = self.client.client_guid

        if smb2.DIALECT_SMB3_1_1 in neg_req.dialects:
            if ciphers is None:
                ciphers = [crypto.SMB2_AES_128_GCM, crypto.SMB2_AES_128_CCM]
            if ciphers:
                encryption_req = crypto.EncryptionCapabilitiesRequest(neg_req)
                encryption_req.ciphers = ciphers

            preauth_integrity_req = smb2.PreauthIntegrityCapabilitiesRequest(neg_req)
            if hash_algorithms is None:
                hash_algorithms = [smb2.SMB2_SHA_512]
            preauth_integrity_req.hash_algorithms = hash_algorithms
            if salt is not None:
                preauth_integrity_req.salt = salt
            else:
                preauth_integrity_req.salt = array.array(
                    "B", map(random.randint, [0] * 32, [255] * 32)
                )
        self._negotiate_request = neg_req
        return neg_req

    def negotiate_submit(self, negotiate_request):
        negotiate_future = self.submit(negotiate_request.parent.parent)[0]

        def assign_response(f):
            self.negotiate_response = f.result()[0]

        negotiate_future.then(assign_response)
        return negotiate_future

    def negotiate(self, hash_algorithms=None, salt=None, ciphers=None):
        """
        Perform dialect negotiation.

        This must be performed before setting up a session with
        :py:func:`Connection.session_setup`.
        """
        self.negotiate_submit(
            self.negotiate_request(hash_algorithms, salt, ciphers)
        ).result()
        return self

    class SessionSetupContext(object):
        def __init__(self, conn, creds=None, bind=None, resume=None, ntlm_version=None):
            assert conn.negotiate_response is not None

            self.conn = conn
            self.dialect_revision = conn.negotiate_response.dialect_revision
            self.bind = bind
            self.resume = resume
            self._pre_auth_integrity_hash = conn._pre_auth_integrity_hash[:]

            if creds and auth.ntlm is not None:
                self.auth = auth.NtlmProvider(conn, creds)
                if ntlm_version is not None:
                    self.auth.authenticator.ntlm_version = ntlm_version
            elif auth.kerberos is not None:
                self.auth = auth.KerberosProvider(conn, creds)
            else:
                raise ImportError(
                    "Neither ntlm nor kerberos authentication methods are available"
                )

            self._settings = {}
            self.prev_session_id = 0
            self.session_id = 0
            self.requests = []
            self.responses = []
            self.session_future = Future(request=self.requests)
            self.interim_future = None

            if bind:
                assert conn.negotiate_response.dialect_revision >= 0x300
                self.session_id = bind.session_id
                conn._binding = bind
                # assume the signing key from the previous session
                conn._binding_key = bind.first_channel().signing_key
            elif resume:
                assert conn.negotiate_response.dialect_revision >= 0x300
                self.prev_session_id = resume.session_id

        def let(self, **kwargs):
            return core.Let(self, kwargs)

        def derive_signing_key(self, session_key=None, context=None):
            if session_key is None:
                session_key = self.session_key
            if self.dialect_revision >= smb2.DIALECT_SMB3_1_1:
                if context is None:
                    context = self._pre_auth_integrity_hash
                return digest.derive_key(session_key, b"SMBSigningKey", context)[:16]
            elif self.dialect_revision >= smb2.DIALECT_SMB3_0:
                if context is None:
                    context = b"SmbSign\0"
                return digest.derive_key(session_key, b"SMB2AESCMAC", context)[:16]
            else:
                return session_key

        def derive_encryption_keys(self, session_key=None, context=None):
            if self.dialect_revision >= smb2.DIALECT_SMB3_1_1:
                if context is None:
                    context = self._pre_auth_integrity_hash
                for nctx in self.conn.negotiate_response:
                    if isinstance(nctx, crypto.EncryptionCapabilitiesResponse):
                        try:
                            return crypto.EncryptionContext(
                                crypto.CryptoKeys311(self.session_key, context),
                                nctx.ciphers,
                            )
                        except crypto.CipherMismatch:
                            pass
            elif self.dialect_revision >= smb2.DIALECT_SMB3_0:
                if (
                    self.conn.negotiate_response.capabilities
                    & smb2.SMB2_GLOBAL_CAP_ENCRYPTION
                ):
                    return crypto.EncryptionContext(
                        crypto.CryptoKeys300(self.session_key),
                        [crypto.SMB2_AES_128_CCM],
                    )

        def _update_pre_auth_integrity(self, packet, data=None):
            if smb2.DIALECT_SMB3_1_1 not in self.conn.client.dialects:
                # hash only applies if client requests 3.1.1
                return
            neg_resp = self.conn.negotiate_response
            if (
                neg_resp is not None
                and neg_resp.dialect_revision < smb2.DIALECT_SMB3_1_1
            ):
                # hash only applies if server negotiates 3.1.1
                return
            if data is None:
                data = packet.serialize()
            self._pre_auth_integrity_hash = digest.smb3_sha512(
                self._pre_auth_integrity_hash + data
            )

        def _send_session_setup(self, sec_buf):
            smb_req = self.conn.request()
            session_req = smb2.SessionSetupRequest(smb_req)

            smb_req.session_id = self.session_id
            session_req.previous_session_id = self.prev_session_id
            session_req.security_mode = smb2.SMB2_NEGOTIATE_SIGNING_ENABLED
            session_req.security_buffer = sec_buf
            if self.bind:
                smb_req.flags = smb2.SMB2_FLAGS_SIGNED
                session_req.flags = smb2.SMB2_SESSION_FLAG_BINDING

            for (attr, value) in self._settings.items():
                setattr(session_req, attr, value)

            self.requests.append(smb_req)
            return self.conn.submit(smb_req.parent)[0]

        def _finish(self, smb_res):
            sec_buf = smb_res[0].security_buffer
            out_buf, self.session_key = self.auth.step(sec_buf)
            signing_key = self.derive_signing_key()
            encryption_context = self.derive_encryption_keys()

            # Verify final signature
            smb_res.verify(self.conn.signing_digest(), signing_key)

            if self.bind:
                self.conn._binding = None
                self.conn._binding_key = None
                session = self.bind
            else:
                session = Session(
                    self.conn.client,
                    self.session_id,
                    self.session_key,
                    encryption_context,
                    smb_res,
                )
                session.user = self.auth.username()

            return session.addchannel(self.conn, signing_key)

        def __iter__(self):
            return self

        def submit(self, f=None):
            """
            Submit rounds of SessionSetupRequests

            Returns a L{Future} object, for the L{Channel} object
            """
            try:
                res = next(self)
                res.then(self.submit)
            except StopIteration:
                pass
            return self.session_future

        def __next__(self):
            with self.session_future:
                res = self._process()
                if res is not None:
                    return res
            raise StopIteration()

        def _process(self):
            out_buf = None
            if not self.interim_future and not self.responses:
                # send the initial request
                out_buf, self.session_key = self.auth.step(
                    self.conn.negotiate_response.security_buffer
                )

            elif self.interim_future:
                # handle pre-auth integrity on the previous request
                previous_request = self.requests[-1]
                self._update_pre_auth_integrity(previous_request)

                smb_res = self.interim_future.result()
                self.interim_future = None
                self.responses.append(smb_res)
                self.session_id = smb_res.session_id

                if smb_res.status == ntstatus.STATUS_SUCCESS:
                    # session is established
                    with self.session_future:
                        self.session_future(self._finish(smb_res))
                    return self.session_future
                else:
                    # process interim request
                    self._update_pre_auth_integrity(smb_res, smb_res.parent.buf[4:])
                    session_res = smb_res[0]
                    if self.bind:
                        # Need to verify intermediate signatures
                        smb_res.verify(
                            self.conn.signing_digest(), self.conn._binding_key
                        )
                    out_buf, self.session_key = self.auth.step(
                        session_res.security_buffer
                    )
            if out_buf:
                # submit additional requests if necessary
                self.interim_future = self._send_session_setup(out_buf)
                return self.interim_future

    def session_setup(self, creds=None, bind=None, resume=None):
        """
        Establish a session.

        Establishes a session, performing GSS rounds as necessary.  Returns
        a L{Channel} object which can be used for further requests on the given
        connection and session.

        @type creds: str
        @param creds: A set of credentials of the form '<domain>\<user>%<password>'.
                      If specified, NTLM authentication will be used.  If None,
                      Kerberos authentication will be attempted.
        @type bind: L{Session}
        @param bind: An existing session to bind.
        @type resume: L{Session}
        @param resume: An previous session to resume.
        """
        session_context = self.SessionSetupContext(self, creds, bind, resume)
        return session_context.submit().result()

    # Return a fresh netbios frame with connection as context
    def frame(self):
        return netbios.Netbios(context=self)

    # Return a fresh smb2 frame with connection as context
    # Put it in a netbios frame automatically if none given
    def request(self, parent=None):
        if parent is None:
            parent = self.frame()
        req = smb2.Smb2(parent, context=self)
        req.channel_sequence = self.client.channel_sequence

        for (attr, value) in self._settings.items():
            setattr(req, attr, value)

        return req

    def let(self, **kwargs):
        return core.Let(self, kwargs)

    #
    # SMB2 context upcalls
    #
    def session(self, session_id):
        return self._sessions.get(session_id, None)

    def signing_key(self, session_id):
        if session_id in self._sessions:
            session = self._sessions[session_id]
            channel = session._channels[id(self)]
            return channel.signing_key
        elif self._binding and self._binding.session_id == session_id:
            return self._binding_key

    def encryption_context(self, session_id):
        if session_id in self._sessions:
            session = self._sessions[session_id]
            return session.encryption_context

    def signing_digest(self):
        assert self.negotiate_response is not None
        if self.negotiate_response.dialect_revision >= smb2.DIALECT_SMB3_0:
            return digest.aes128_cmac
        else:
            return digest.sha256_hmac

    def get_request(self, message_id):
        if message_id in self._future_map:
            return self._future_map[message_id].request
        else:
            return None


class Session(object):
    """
    Represents an SMB2 session.

    May contain one or more active channels (connections) or trees
    """
    def __init__(self, client, session_id, session_key, encryption_context, smb_res):
        self.client = client
        self.session_id = session_id
        self.session_key = session_key
        self.encryption_context = encryption_context
        self.encrypt_data = False
        if (
            smb_res[0].session_flags & smb2.SMB2_SESSION_FLAG_ENCRYPT_DATA
            and self.encryption_context is not None
        ):
            self.encrypt_data = True
        self._channels = {}
        self._trees = {}
        self.user = None

    def addchannel(self, conn, signing_key):
        channel = Channel(conn, self, signing_key)
        self._channels[id(conn)] = channel
        conn._sessions[self.session_id] = self
        return channel

    def delchannel(self, conn):
        del self._channels[id(conn)]

    def first_channel(self):
        return next(iter(self._channels.values()))

    def tree(self, tree_id):
        return self._trees.get(tree_id, None)


class Channel(object):
    """
    SMB2 communication channel associated with an underlying
    :py:class:`Connection` and :py:class:`Session`.

    All post-authentication protocol commands have blocking and non-blocking
    helper methods defined in this class.
    """
    def __init__(self, connection, session, signing_key):
        object.__init__(self)
        self.connection = connection
        self.session = session
        self.signing_key = signing_key

    def cancel_request(self, future):
        if future.response is not None:
            raise StateError("Cannot cancel completed request")

        smb_req = self.request()
        cancel_req = smb2.Cancel(smb_req)

        # Don't bother trying to sign cancel
        smb_req.flags &= ~smb2.SMB2_FLAGS_SIGNED

        # Use async id to cancel if applicable:
        if future.interim_response is not None:
            smb_req.async_id = future.interim_response.async_id
            smb_req.tree_id = None
            smb_req.flags |= smb2.SMB2_FLAGS_ASYNC_COMMAND
            smb_req.message_id = 0
        else:
            smb_req.message_id = future.request.message_id

        return cancel_req

    def cancel(self, future):
        cancel_req = self.cancel_request(future)
        return self.connection.submit(cancel_req.parent.parent)[0]

    def tree_connect_request(self, path):
        smb_req = self.request()
        if self.connection.negotiate_response.dialect_revision >= smb2.DIALECT_SMB3_1_1:
            smb_req.flags |= smb2.SMB2_FLAGS_SIGNED
        tree_req = smb2.TreeConnectRequest(smb_req)
        tree_req.path = "\\\\" + self.connection.server + "\\" + path
        return tree_req

    def tree_connect_submit(self, tree_req):
        tree_future = Future(request=tree_req.parent)
        resp_future = self.connection.submit(tree_req.parent.parent)[0]
        resp_future.then(
            lambda f: tree_future.complete(
                Tree(self.session, tree_req.path, f.result())
            )
        )
        return tree_future

    def tree_connect(self, path):
        return self.tree_connect_submit(self.tree_connect_request(path)).result()

    def tree_disconnect_request(self, tree):
        smb_req = self.request(obj=tree)
        tree_req = smb2.TreeDisconnectRequest(smb_req)
        return tree_req

    def tree_disconnect(self, tree):
        return self.connection.transceive(
            self.tree_disconnect_request(tree).parent.parent
        )[0]

    def logoff_request(self):
        smb_req = self.request()
        logoff_req = smb2.LogoffRequest(smb_req)
        return logoff_req

    def logoff_submit(self, logoff_req):
        def logoff_finish(f):
            for channel in self.session._channels.values():
                del channel.connection._sessions[self.session.session_id]

        logoff_future = self.connection.submit(logoff_req.parent.parent)[0]
        logoff_future.then(logoff_finish)
        return logoff_future

    def logoff(self):
        return self.logoff_submit(self.logoff_request()).result()

    def create_request(
        self,
        tree,
        path,
        access=smb2.GENERIC_READ | smb2.GENERIC_WRITE,
        attributes=smb2.FILE_ATTRIBUTE_NORMAL,
        share=0,
        disposition=smb2.FILE_OPEN_IF,
        options=0,
        maximal_access=None,
        oplock_level=smb2.SMB2_OPLOCK_LEVEL_NONE,
        lease_key=None,
        lease_state=None,
        lease_flags_v2=None,
        durable=False,
        persistent=False,
        create_guid=None,
        app_instance_id=None,
        query_on_disk_id=False,
        extended_attributes=None,
        timewarp=None,
    ):

        prev_open = None

        smb_req = self.request(obj=tree)
        create_req = smb2.CreateRequest(smb_req)

        create_req.name = path
        create_req.desired_access = access
        create_req.file_attributes = attributes
        create_req.share_access = share
        create_req.create_disposition = disposition
        create_req.create_options = options
        create_req.requested_oplock_level = oplock_level

        if maximal_access:
            max_req = smb2.MaximalAccessRequest(create_req)
            if maximal_access is not True:
                max_req.timestamp = maximal_access

        if oplock_level == smb2.SMB2_OPLOCK_LEVEL_LEASE:
            if lease_key is None:
                lease_key = crypto.random_bytes(16)
            if lease_state is None:
                lease_state = smb2.SMB2_LEASE_RWH
            lease_req = smb2.LeaseRequest(create_req, lease_flags_v2)
            lease_req.lease_key = lease_key
            lease_req.lease_state = lease_state

        if isinstance(durable, pike.io.Open):
            prev_open = durable
            if durable.durable_timeout is None:
                durable_req = smb2.DurableHandleReconnectRequest(create_req)
                durable_req.file_id = durable.file_id
            else:
                durable_req = smb2.DurableHandleReconnectV2Request(create_req)
                durable_req.file_id = durable.file_id
                durable_req.create_guid = durable.create_guid
                durable_req.flags = durable.durable_flags
        elif durable is True:
            durable_req = smb2.DurableHandleRequest(create_req)
        elif durable is not False:
            durable_req = smb2.DurableHandleV2Request(create_req)
            durable_req.timeout = durable
            if persistent:
                durable_req.flags = smb2.SMB2_DHANDLE_FLAG_PERSISTENT
            if create_guid is None:
                create_guid = array.array(
                    "B", map(random.randint, [0] * 16, [255] * 16)
                )
            durable_req.create_guid = create_guid

        if app_instance_id:
            app_instance_id_req = smb2.AppInstanceIdRequest(create_req)
            app_instance_id_req.app_instance_id = app_instance_id

        if query_on_disk_id:
            query_on_disk_id_req = smb2.QueryOnDiskIDRequest(create_req)

        if extended_attributes:
            ext_attr_len = len(extended_attributes)
            for name, value in extended_attributes.items():
                ext_attr = smb2.ExtendedAttributeRequest(create_req)
                if ext_attr_len == 1:
                    next_entry_offset = 0
                else:
                    next_entry_offset = 10 + len(name) + len(value)
                ext_attr.next_entry_offset = next_entry_offset
                ext_attr.ea_name = name
                ext_attr.ea_name_length = len(name)
                ext_attr.ea_value = value
                ext_attr.ea_value_length = len(value)
                ext_attr_len = ext_attr_len - 1

        if timewarp:
            timewarp_req = smb2.TimewarpTokenRequest(create_req)
            timewarp_req.timestamp = nttime.NtTime(timewarp)

        open_future = Future(request=create_req.parent)

        def finish(f):
            with open_future:
                open_future(
                    pike.io.Open(
                        tree,
                        create_request=create_req,
                        create_response=f.result()[0],
                        create_guid=create_guid,
                        previous_open=prev_open,
                    )
                )

        create_req.open_future = open_future
        create_req.finish = finish

        return create_req

    def create_submit(self, create_req):
        open_future = create_req.open_future
        open_future.request_future = self.connection.submit(create_req.parent.parent)[0]
        open_future.request_future.then(create_req.finish)

        return open_future

    def create(
        self,
        tree,
        path,
        access=smb2.GENERIC_READ | smb2.GENERIC_WRITE,
        attributes=smb2.FILE_ATTRIBUTE_NORMAL,
        share=0,
        disposition=smb2.FILE_OPEN_IF,
        options=0,
        maximal_access=None,
        oplock_level=smb2.SMB2_OPLOCK_LEVEL_NONE,
        lease_key=None,
        lease_state=None,
        lease_flags_v2=None,
        durable=False,
        persistent=False,
        create_guid=None,
        app_instance_id=None,
        query_on_disk_id=False,
        extended_attributes=None,
        timewarp=None,
    ):
        return self.create_submit(
            self.create_request(
                tree,
                path,
                access,
                attributes,
                share,
                disposition,
                options,
                maximal_access,
                oplock_level,
                lease_key,
                lease_state,
                lease_flags_v2,
                durable,
                persistent,
                create_guid,
                app_instance_id,
                query_on_disk_id,
                extended_attributes,
                timewarp,
            )
        )

    def close_request(self, handle):
        smb_req = self.request(obj=handle)
        close_req = smb2.CloseRequest(smb_req)

        close_req.file_id = handle.file_id
        close_req.handle = handle
        return close_req

    def close_submit(self, close_req):
        resp_future = self.connection.submit(close_req.parent.parent)[0]
        resp_future.then(lambda f: close_req.handle.dispose())
        return resp_future

    def close(self, handle):
        return self.close_submit(self.close_request(handle)).result()

    def query_directory_request(
        self,
        handle,
        file_information_class=smb2.FILE_DIRECTORY_INFORMATION,
        flags=0,
        file_index=0,
        file_name="*",
        output_buffer_length=8192,
    ):
        smb_req = self.request(obj=handle)
        enum_req = smb2.QueryDirectoryRequest(smb_req)
        enum_req.file_id = handle.file_id
        enum_req.file_name = file_name
        enum_req.output_buffer_length = output_buffer_length
        enum_req.file_information_class = file_information_class
        enum_req.flags = flags
        enum_req.file_index = file_index
        return enum_req

    def query_directory(
        self,
        handle,
        file_information_class=smb2.FILE_DIRECTORY_INFORMATION,
        flags=0,
        file_index=0,
        file_name="*",
        output_buffer_length=8192,
    ):
        return self.connection.transceive(
            self.query_directory_request(
                handle,
                file_information_class,
                flags,
                file_index,
                file_name,
                output_buffer_length,
            ).parent.parent
        )[0][0]

    def enum_directory(
        self,
        handle,
        file_information_class=smb2.FILE_DIRECTORY_INFORMATION,
        file_name="*",
        output_buffer_length=8192,
    ):
        while True:
            try:
                for info in self.query_directory(
                    handle,
                    file_information_class=file_information_class,
                    file_name=file_name,
                    output_buffer_length=output_buffer_length,
                ):
                    yield info
            except ResponseError as e:
                if e.response.status == ntstatus.STATUS_NO_MORE_FILES:
                    return
                else:
                    raise

    def query_file_info_request(
        self,
        create_res,
        file_information_class=smb2.FILE_BASIC_INFORMATION,
        info_type=smb2.SMB2_0_INFO_FILE,
        output_buffer_length=4096,
        additional_information=None,
    ):
        smb_req = self.request(obj=create_res)
        query_req = smb2.QueryInfoRequest(smb_req)

        query_req.info_type = info_type
        query_req.file_information_class = file_information_class
        query_req.file_id = create_res.file_id
        query_req.output_buffer_length = output_buffer_length
        if additional_information:
            query_req.additional_information = additional_information
        return query_req

    def query_file_info(
        self,
        create_res,
        file_information_class=smb2.FILE_BASIC_INFORMATION,
        info_type=smb2.SMB2_0_INFO_FILE,
        output_buffer_length=4096,
        additional_information=None,
        first_result_only=None,
    ):
        if first_result_only is None:
            warnings.warn(
                "In a future release, query_file_info will return QueryInfoResponse "
                "instead of the first result from QueryInfoResponse. To maintain "
                "the previous behavior (and silence this warning) pass "
                "`first_result_only=True`. To opt-in to the new behavior, pass "
                "`first_result_only=False`, which will become the new default.",
                DeprecationWarning,
                stacklevel=2,
            )
            first_result_only = True
        resp = self.connection.transceive(
            self.query_file_info_request(
                create_res,
                file_information_class,
                info_type,
                output_buffer_length,
                additional_information,
            ).parent.parent
        )[0][0]
        if first_result_only:
            # only returning the first child info
            return resp[0]
        return resp

    def set_file_info_request(
        self,
        handle,
        file_information_class=smb2.FILE_BASIC_INFORMATION,
        info_type=smb2.SMB2_0_INFO_FILE,
        input_buffer_length=4096,
        additional_information=None,
    ):
        smb_req = self.request(obj=handle)
        set_req = smb2.SetInfoRequest(smb_req)
        set_req.file_id = handle.file_id
        set_req.file_information_class = file_information_class
        set_req.info_type = info_type
        set_req.input_buffer_length = input_buffer_length
        if additional_information:
            set_req.additional_information = additional_information
        return set_req

    @contextlib.contextmanager
    def set_file_info(self, handle, cls, additional_information=None):
        info_type = file_information_class = None
        if hasattr(cls, "info_type"):
            info_type = cls.info_type
        if hasattr(cls, "file_information_class"):
            file_information_class = cls.file_information_class
        set_req = self.set_file_info_request(
            handle,
            file_information_class,
            info_type,
            additional_information=additional_information,
        )
        yield cls(set_req)
        self.connection.transceive(set_req.parent.parent)[0]

    def query_quota_info_request(
        self, handle, return_single=False, restart_scan=False, output_buffer_length=4096
    ):
        smb_req = self.request(obj=handle)
        query_req = smb2.QueryInfoRequest(smb_req)
        quota_query_info = smb2.QueryQuotaInfo(query_req)

        query_req.info_type = smb2.SMB2_0_INFO_QUOTA
        query_req.file_information_class = 0
        query_req.file_id = handle.file_id
        query_req.output_buffer_length = output_buffer_length
        quota_query_info.return_single = return_single
        quota_query_info.restart_scan = restart_scan
        return query_req

    def query_quota_info(
        self, handle, return_single=False, restart_scan=False, output_buffer_length=4096
    ):
        return self.connection.transceive(
            self.query_quota_info_request(
                handle,
                return_single=return_single,
                restart_scan=restart_scan,
                output_buffer_length=output_buffer_length,
            ).parent.parent
        )[0][0]

    def change_notify_request(
        self,
        handle,
        completion_filter=smb2.SMB2_NOTIFY_CHANGE_CREATION,
        flags=0,
        buffer_length=4096,
    ):
        smb_req = self.request(obj=handle)
        cnotify_req = smb2.ChangeNotifyRequest(smb_req)
        cnotify_req.file_id = handle.file_id
        cnotify_req.buffer_length = buffer_length
        cnotify_req.flags = flags
        cnotify_req.completion_filter = completion_filter
        return cnotify_req

    def change_notify(
        self,
        handle,
        completion_filter=smb2.SMB2_NOTIFY_CHANGE_CREATION,
        flags=0,
        buffer_length=4096,
    ):
        return self.connection.submit(
            self.change_notify_request(
                handle,
                completion_filter,
                flags,
                buffer_length=buffer_length,
            ).parent.parent
        )[0]

    # Send an echo request and get a response
    def echo(self):
        # Create request structure
        smb_req = self.request()
        # Make the request struct have an ECHO_REQUEST
        enum_req = smb2.EchoRequest(smb_req)
        # Get response  first [0] = first response, 2nd [0] = echo response
        # frame
        return self.connection.transceive(smb_req.parent)[0][0]

    def flush_request(self, file):
        smb_req = self.request(obj=file)
        flush_req = smb2.FlushRequest(smb_req)
        flush_req.file_id = file.file_id
        return flush_req

    def flush(self, file):
        return self.connection.transceive(self.flush_request(file).parent.parent)[0]

    def read_request(self, file, length, offset, minimum_count=0, remaining_bytes=0):
        smb_req = self.request(obj=file)
        read_req = smb2.ReadRequest(smb_req)

        read_req.length = length
        read_req.offset = offset
        read_req.minimum_count = minimum_count
        read_req.remaining_bytes = remaining_bytes
        read_req.file_id = file.file_id
        return read_req

    def read(self, file, length, offset, minimum_count=0, remaining_bytes=0):
        return self.connection.transceive(
            self.read_request(
                file, length, offset, minimum_count, remaining_bytes
            ).parent.parent
        )[0][0].data

    def write_request(self, file, offset, buffer=None, remaining_bytes=0, flags=0):
        """
        Create a pike.smb2.WriteRequest from the given parameters

        @param file: L{Open}
        @param offset: int offset into the file
        @param buffer: bytes or array.array('B'). If a unicode str is passed, it
            can only contain ascii characters and a warning will be raised.
        @param remaining_bytes:
        @param flags: L{pike.smb2.WriteFlags}
        """
        if isinstance(buffer, array.array) and buffer.typecode != "B":
            raise ValueError(
                "array.array must have typecode 'B', not {!r}".format(buffer.typecode)
            )
        elif isinstance(buffer, str):
            warnings.warn(
                "buffer must be bytes, got {!r}, casting as str and encoding "
                "with 'ascii'".format(type(buffer)),
                UnicodeWarning,
            )
            buffer = buffer.encode("ascii")
        if sys.version_info < (3,) and isinstance(buffer, memoryview):
            # explicit bytes-cast required on py2
            buffer = buffer.tobytes()
        if buffer is not None and not isinstance(
            buffer, (array.array, bytes, memoryview, bytearray)
        ):
            raise TypeError(
                "buffer must be a byte string or byte array, not {!r}".format(
                    type(buffer)
                )
            )
        smb_req = self.request(obj=file)
        write_req = smb2.WriteRequest(smb_req)

        write_req.offset = offset
        write_req.file_id = file.file_id
        write_req.buffer = buffer
        write_req.remaining_bytes = remaining_bytes
        write_req.flags = flags
        return write_req

    def write(self, file, offset, buffer=None, remaining_bytes=0, flags=0):
        smb_res = self.connection.transceive(
            self.write_request(
                file, offset, buffer, remaining_bytes, flags
            ).parent.parent
        )

        return smb_res[0][0].count

    def lock_request(self, handle, locks, sequence=0):
        """
        @param locks: A list of lock tuples, each of which consists of (offset, length, flags).
        """
        smb_req = self.request(obj=handle)
        lock_req = smb2.LockRequest(smb_req)

        lock_req.file_id = handle.file_id
        lock_req.locks = locks
        lock_req.lock_sequence = sequence
        return lock_req

    def lock(self, handle, locks, sequence=0):
        """
        @param locks: A list of lock tuples, each of which consists of (offset, length, flags).
        """
        return self.connection.submit(
            self.lock_request(handle, locks, sequence).parent.parent
        )[0]

    def validate_negotiate_info(self, tree):
        smb_req = self.request(obj=tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        vni_req = smb2.ValidateNegotiateInfoRequest(ioctl_req)
        client = self.session.client

        # Validate negotiate must always be signed
        smb_req.flags |= smb2.SMB2_FLAGS_SIGNED
        ioctl_req.flags = smb2.SMB2_0_IOCTL_IS_FSCTL
        vni_req.capabilities = client.capabilities
        vni_req.client_guid = client.client_guid
        vni_req.security_mode = client.security_mode
        vni_req.dialects = client.dialects

        res = self.connection.transceive(smb_req.parent)[0]

        return res

    def query_network_interface_info_request(self, tree):
        smb_req = self.request(obj=tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        qni_req = smb2.QueryNetworkInterfaceInfoRequest(ioctl_req)
        client = self.session.client

        # Windows sends 65536 buffer
        ioctl_req.max_output_response = 65536
        ioctl_req.flags = smb2.SMB2_0_IOCTL_IS_FSCTL

        return ioctl_req

    def query_network_interface_info(self, tree):
        return self.connection.transceive(
            self.query_network_interface_info_request(tree).parent.parent
        )[0]

    def resume_key(self, file):
        smb_req = self.request(obj=file.tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        resumekey_req = smb2.RequestResumeKeyRequest(ioctl_req)

        ioctl_req.file_id = file.file_id
        ioctl_req.flags |= smb2.SMB2_0_IOCTL_IS_FSCTL

        return self.connection.transceive(smb_req.parent)[0]

    def network_resiliency_request_request(self, file, timeout):
        smb_req = self.request(obj=file.tree)
        ioctl_req = smb2.IoctlRequest(smb_req)

        nrr_req = smb2.NetworkResiliencyRequestRequest(ioctl_req)
        ioctl_req.file_id = file.file_id
        ioctl_req.max_output_response = 4096
        ioctl_req.flags = smb2.SMB2_0_IOCTL_IS_FSCTL
        nrr_req.timeout = timeout
        nrr_req.reserved = 0
        return ioctl_req

    def network_resiliency_request(self, file, timeout):
        def update_handle(resp_future):
            if resp_future.result().status == ntstatus.STATUS_SUCCESS:
                # 3.3.5.15.9 Handling a Resiliency Request
                file._is_durable = False
                file._is_resilient = True

        nrr_future = self.connection.submit(
            self.network_resiliency_request_request(file, timeout).parent.parent
        )[0]
        nrr_future.then(update_handle)
        return nrr_future.result()

    def copychunk_request(
        self, source_file, target_file, chunks, resume_key=None, write_flag=False
    ):
        """
        @param source_file: L{Open}
        @param target_file: L{Open}
        @param chunks: sequence of tuples (source_offset, target_offset, length)
        """
        if not resume_key:
            resume_key = self.resume_key(source_file)[0][0].resume_key

        smb_req = self.request(obj=target_file.tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        if write_flag:
            copychunk_req = smb2.CopyChunkCopyWriteRequest(ioctl_req)
        else:
            copychunk_req = smb2.CopyChunkCopyRequest(ioctl_req)

        ioctl_req.max_output_response = 12
        ioctl_req.file_id = target_file.file_id
        ioctl_req.flags |= smb2.SMB2_0_IOCTL_IS_FSCTL
        copychunk_req.source_key = resume_key
        copychunk_req.chunk_count = len(chunks)

        for source_offset, target_offset, length in chunks:
            chunk = smb2.CopyChunk(copychunk_req)
            chunk.source_offset = source_offset
            chunk.target_offset = target_offset
            chunk.length = length
        return ioctl_req

    def copychunk(
        self, source_file, target_file, chunks, resume_key=None, write_flag=False
    ):
        """
        @param source_file: L{Open}
        @param target_file: L{Open}
        @param chunks: sequence of tuples (source_offset, target_offset, length)
        """
        return self.connection.transceive(
            self.copychunk_request(
                source_file, target_file, chunks, resume_key, write_flag
            ).parent.parent
        )[0]

    def set_symlink_request(self, file, target_name, flags):
        smb_req = self.request(obj=file.tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        set_reparse_req = smb2.SetReparsePointRequest(ioctl_req)
        symlink_buffer = smb2.SymbolicLinkReparseBuffer(set_reparse_req)

        ioctl_req.max_output_response = 0
        ioctl_req.file_id = file.file_id
        ioctl_req.flags |= smb2.SMB2_0_IOCTL_IS_FSCTL
        symlink_buffer.substitute_name = target_name
        symlink_buffer.flags = flags
        return ioctl_req

    def set_symlink(self, file, target_name, flags):
        return self.connection.transceive(
            self.set_symlink_request(file, target_name, flags).parent.parent
        )[0]

    def get_symlink_request(self, file):
        smb_req = self.request(obj=file.tree)
        ioctl_req = smb2.IoctlRequest(smb_req)
        set_reparse_req = smb2.GetReparsePointRequest(ioctl_req)

        ioctl_req.file_id = file.file_id
        ioctl_req.flags |= smb2.SMB2_0_IOCTL_IS_FSCTL
        return ioctl_req

    def get_symlink(self, file):
        return self.connection.transceive(self.get_symlink_request(file).parent.parent)[
            0
        ]

    def fsctl_request(self, smb_req, fh, buflen=16384):
        ioctl_req = smb2.IoctlRequest(smb_req)
        ioctl_req.max_output_response = buflen
        ioctl_req.flags = smb2.SMB2_0_IOCTL_IS_FSCTL
        ioctl_req.file_id = fh.file_id
        return ioctl_req

    def enumerate_snapshots_request(
        self, fh, snap_request=smb2.EnumerateSnapshotsRequest, max_output_response=16384
    ):
        smb_req = self.request(obj=fh.tree)
        fsctl_req = self.fsctl_request(smb_req, fh, buflen=max_output_response)
        enum_req = snap_request(fsctl_req)
        return enum_req

    def enumerate_snapshots(self, fh, snap_request=smb2.EnumerateSnapshotsRequest):
        return self.connection.transceive(
            self.enumerate_snapshots_request(fh, snap_request).parent.parent.parent
        )[0]

    def enumerate_snapshots_list(self, fh, snap_request=smb2.EnumerateSnapshotsRequest):
        return self.enumerate_snapshots(fh, snap_request)[0][0].snapshots

    def zero_data(self, tree, src_offsets, dst_offsets, src_filename):
        """Send a FSCTL_SET_ZERO_DATA ioctl request"""
        fh_src = self.create(
            tree,
            src_filename,
            access=smb2.FILE_READ_DATA | smb2.FILE_WRITE_DATA,
            share=(
                smb2.FILE_SHARE_READ | smb2.FILE_SHARE_WRITE | smb2.FILE_SHARE_DELETE
            ),
            disposition=smb2.FILE_OPEN_IF,
        ).result()
        smb_req = self.request(obj=tree)
        fsctl_req = self.fsctl_request(smb_req, fh_src)
        smb2.SetSparseRequest(fsctl_req)
        try:
            results = self.connection.transceive(smb_req.parent)
        except Exception:
            self.close(fh_src)
            raise

        smb_req = self.request(obj=tree)
        fsctl_req = self.fsctl_request(smb_req, fh_src)
        zerodata_req = smb2.SetZeroDataRequest(fsctl_req)
        zerodata_req.file_offset = src_offsets
        zerodata_req.beyond_final_zero = dst_offsets
        try:
            results = self.connection.transceive(smb_req.parent)
            return results
        finally:
            self.close(fh_src)

    def lease_break_acknowledgement(self, tree, notify):
        """
        @param tree: L{Tree} which the lease is taken against
        @param notify: L{Smb2} frame containing a L{LeaseBreakRequest}
        @return: a L{LeaseBreakAcknowledgement} with some fields pre-populated
        """
        lease_break = notify[0]
        smb_req = self.request(obj=tree)
        ack_req = smb2.LeaseBreakAcknowledgement(smb_req)
        ack_req.lease_key = lease_break.lease_key
        ack_req.lease_state = lease_break.new_lease_state
        return ack_req

    def oplock_break_acknowledgement(self, fh, notify):
        """
        @param fh: Acknowledge break on this L{Open}
        @param notify: L{Smb2} frame containing a L{OplockBreakRequest}
        @return: an L{OplockBreakAcknowledgement} with some fields pre-populated
        """
        oplock_break = notify[0]
        smb_req = self.request(obj=fh)
        ack_req = smb2.OplockBreakAcknowledgement(smb_req)
        ack_req.file_id = oplock_break.file_id
        ack_req.oplock_level = oplock_break.oplock_level
        return ack_req

    def frame(self):
        return self.connection.frame()

    def request(self, nb=None, obj=None, encrypt_data=None):
        smb_req = self.connection.request(nb)
        smb_req.session_id = self.session.session_id

        if isinstance(obj, Tree):
            smb_req.tree_id = obj.tree_id
        elif isinstance(obj, pike.io.Open):
            smb_req.tree_id = obj.tree.tree_id

        # encryption unspecified, follow session/tree negotiation
        if encrypt_data is None:
            encrypt_data = self.session.encrypt_data
            if isinstance(obj, Tree):
                encrypt_data |= obj.encrypt_data
            elif isinstance(obj, pike.io.Open):
                encrypt_data |= obj.tree.encrypt_data

        # a packet is either encrypted or signed
        if encrypt_data and self.session.encryption_context is not None:
            transform = crypto.TransformHeader(smb_req.parent)
            transform.encryption_context = self.session.encryption_context
            transform.session_id = self.session.session_id
        elif (
            self.connection.negotiate_response.security_mode
            & smb2.SMB2_NEGOTIATE_SIGNING_REQUIRED
            or self.connection.client.security_mode
            & smb2.SMB2_NEGOTIATE_SIGNING_REQUIRED
        ):
            smb_req.flags |= smb2.SMB2_FLAGS_SIGNED

        return smb_req

    def let(self, **kwargs):
        return self.connection.let(**kwargs)


class Tree(object):
    """
    An SMB2 Tree object is associated with a :py:class:`Session`.

    A Tree can be used as a contextmanager to automatically call
    :py:func:`~Channel.tree_disconnect` after exiting the context block.

    Using the division operator (``/``) against a :py:class:`Tree` returns a
    :py:class:`~pike.path.PikePath` that can be used to perform ``Path``
    operations on the share
    """
    def __init__(self, session, path, smb_res):
        object.__init__(self)
        self.session = session
        self.path = path
        self.tree_id = smb_res.tree_id
        self.tree_connect_response = smb_res[0]
        self.encrypt_data = False
        if smb_res[0].share_flags & smb2.SMB2_SHAREFLAG_ENCRYPT_DATA:
            self.encrypt_data = True
        self.session._trees[self.tree_id] = self

    def close(self):
        if self.tree_id is not None:
            self.session.first_channel().tree_disconnect(self)
            self.tree_id = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __fspath__(self):
        """
        Return the string path representation of the Tree.
        """
        return r"\\{}\{}".format(
            self.session.first_channel().connection.hostname,
            self.path.rpartition("\\")[2],
        )

    def __add__(self, other):
        """
        Concatenate from the string representation of the Tree.
        """
        return self.__fspath__() + other

    def __truediv__(self, key):
        """
        Return a PikePath with the given key joined.
        """
        from . import path

        return path.PikePath(self, key)

    if sys.version_info < (3,):
        __div__ = __truediv__


class RelatedOpen(object):
    """
    Use in place of a real `Open` object in compound requests to use a handle
    from previous Create in the chain
    """

    def __init__(self, tree=None):
        self.tree_id = tree.tree_id
        self.file_id = smb2.RELATED_FID


class Lease(object):
    """
    SMB2 Lease state attached to a :py:class:`~pike.io.Open` with an active lease.
    """

    def __init__(self, tree):
        self.tree = tree
        self.refs = 1
        self.future = None

    def update(self, lease_res):
        self.lease_key = lease_res.lease_key
        self.lease_state = lease_res.lease_state
        self.lease_epoch = lease_res.epoch
        if self.future is None:
            self.arm_future()

    def arm_future(self):
        """
        (Re)arm the lease future for this Lease. This function should be called
        when a lease changes state to anything other than SMB2_LEASE_NONE
        """
        self.future = self.tree.session.client.lease_break_future(self.lease_key)

    def ref(self):
        self.refs += 1

    def dispose(self):
        self.refs -= 1
        if self.refs == 0:
            self.tree.session.client.dispose_lease(self)

    def on_break(self, cb):
        """
        Simple lease break callback handler.

        :type cb: Callable[[pike.smb2.LeaseState], pike.smb2.LeaseState])
        :param cb: callback function is passed the
            ``pike.smb2.LeaseBreakNotification.new_lease_state`` and must
            return the desired :py:class:`pike.smb2.LeaseState` to break to.
        """

        def simple_handle_break(lease, smb_res, cb_ctx):
            """
            note that lease is not used in this callback,
            since it already closes over self
            """
            notify = smb_res[0]
            self.lease_epoch = notify.new_epoch
            if notify.flags & smb2.SMB2_NOTIFY_BREAK_LEASE_FLAG_ACK_REQUIRED:
                chan = self.tree.session.first_channel()
                ack = chan.lease_break_acknowledgement(self.tree, smb_res)
                ack.lease_state = cb(notify.new_lease_state)
                ack_res = chan.connection.transceive(ack.parent.parent)[0][0]
                if ack_res.lease_state != smb2.SMB2_LEASE_NONE:
                    self.arm_future()
                    self.on_break(cb)
                self.lease_state = ack_res.lease_state
            else:
                self.lease_state = notify.new_lease_state

        self.on_break_request(simple_handle_break)

    def on_break_request(self, cb, cb_ctx=None):
        """
        Complex lease break callback handler.

        :type cb: Callable[[Lease, pike.smb2.Smb2, Optional[Any]], Any])
        :param cb: callback function is passed the lease object, the Smb2
            :py:class:`pike.smb2.LeaseBreakNotification`, and an arbitrary
            context and should handle breaking the lease in some way. The
            callback is also responsible for calling :py:func:`arm_future` and
            and updating the ``.lease_state`` (if changed)
        :type cb_ctx: Optional[Any]
        :param cb_ctx: arbitrary context passed to the callback
        """

        def handle_break(f):
            smb_res = f.result()
            cb(self, smb_res, cb_ctx)

        self.future.then(handle_break)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        set.py
#
# Abstract:
#
#        SMB2_SET_INFO command tests with SMB2_0_INFO_FILE type.
#
# Authors: Rafal Szczesniak (rafal.szczesniak@isilon.com)
#

import datetime

import pike.model
import pike.smb2
import pike.nttime
import pike.test


class SetTest(pike.test.PikeTest):
    def open_file(self):
        self.chan, self.tree = self.tree_connect()
        return self.chan.create(
            self.tree, "test.txt", disposition=pike.smb2.FILE_SUPERSEDE
        ).result()

    # Set timestamps and file attributes with FILE_BASIC_INFORMATION
    def test_set_file_basic_info(self):
        handle = self.open_file()
        now = pike.nttime.NtTime(datetime.datetime.now())

        with self.chan.set_file_info(
            handle, pike.smb2.FileBasicInformation
        ) as file_info:
            file_info.last_write_time = now
            file_info.last_access_time = now

        with self.chan.set_file_info(
            handle, pike.smb2.FileBasicInformation
        ) as file_info:
            file_info.file_attributes = pike.smb2.FILE_ATTRIBUTE_READONLY

        with self.chan.set_file_info(
            handle, pike.smb2.FileBasicInformation
        ) as file_info:
            file_info.file_attributes = pike.smb2.FILE_ATTRIBUTE_NORMAL

        self.chan.close(handle)

    # Set file cursor position with FILE_POSITION_INFORMATION
    def test_set_file_position_info(self):
        handle = self.open_file()

        with self.chan.set_file_info(
            handle, pike.smb2.FilePositionInformation
        ) as file_info:
            file_info.current_byte_offset = 100

        self.chan.close(handle)

    # Set file mode with FILE_MODE_INFORMATION
    def test_set_file_mode_info(self):
        handle = self.open_file()

        with self.chan.set_file_info(
            handle, pike.smb2.FileModeInformation
        ) as file_info:
            file_info.mode = pike.smb2.FILE_SEQUENTIAL_ONLY

        self.chan.close(handle)

    # Set file name with FILE_RENAME_INFORMATION
    def test_set_file_name(self):
        self.chan, self.tree = self.tree_connect()
        handle = self.chan.create(
            self.tree,
            "test.txt",
            disposition=pike.smb2.FILE_SUPERSEDE,
            access=pike.smb2.DELETE,
        ).result()

        with self.chan.set_file_info(
            handle, pike.smb2.FileRenameInformation
        ) as file_info:
            file_info.replace_if_exists = 1
            file_info.file_name = "renamed.txt"

        self.chan.close(handle)

        # open the renamed file for delete and ensure that it exists
        handle2 = self.chan.create(
            self.tree,
            "renamed.txt",
            disposition=pike.smb2.FILE_OPEN,  # fail if doesn't exist
            access=pike.smb2.DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()
        self.chan.close(handle2)



"""
Tests for pike path-like interface
"""
from __future__ import absolute_import, unicode_literals
from builtins import str

import io
import os
import uuid

import pytest

import pike.io
import pike.model as model
import pike.ntstatus as ntstatus
from pike.path import PikePath
import pike.smb2 as smb2


ACCESS_RWD = smb2.GENERIC_READ | smb2.GENERIC_WRITE | smb2.DELETE


def test_stat_root(pike_tmp_path):
    print(pike_tmp_path.stat())


def test_exists(pike_tmp_path):
    unique_id = str(uuid.uuid4())
    fname = pike_tmp_path / unique_id
    assert not fname.exists()
    with pike_tmp_path._channel.create(
        pike_tmp_path._tree,
        fname._path,
        access=ACCESS_RWD,
        options=smb2.FILE_DELETE_ON_CLOSE,
    ).result() as fh:
        assert fname.exists()
    assert not fname.exists()


def test_is_file(pike_tmp_path):
    unique_id = str(uuid.uuid4())
    fname = pike_tmp_path / unique_id
    assert not fname.exists()
    assert not fname.is_file()
    assert not fname.is_symlink()
    fname.touch()
    assert fname.exists()
    assert fname.is_file()
    assert not fname.is_dir()
    assert not fname.is_symlink()
    fname.unlink()


def test_is_dir(pike_tmp_path):
    unique_id = str(uuid.uuid4())
    dname = pike_tmp_path / unique_id
    assert not dname.exists()
    assert not dname.is_dir()
    assert not dname.is_symlink()
    dname.mkdir()
    assert dname.exists()
    assert dname.is_dir()
    assert not dname.is_file()
    assert not dname.is_symlink()
    dname.rmdir()


def test_glob(pike_tmp_path):
    a_files = ["a1", "a2", "a3"]
    b_files = ["b1", "b2", "b3"]
    for f in a_files + b_files:
        (pike_tmp_path / f).touch()

    a_glob = list(pike_tmp_path.glob("a*"))
    assert len(a_glob) == len(a_files)
    for f in a_glob:
        assert f.name in a_files

    b_glob = list(pike_tmp_path.glob("b*"))
    assert len(b_glob) == len(b_files)
    for f in b_glob:
        assert f.name in b_files

    glob2 = list(pike_tmp_path.glob("*2"))
    assert len(glob2) == 2
    for f in glob2:
        assert f.name in ("a2", "b2")


def test_glob_invalid(pike_tmp_path):
    with pytest.raises(ValueError):
        list(pike_tmp_path.glob("**/foo"))

    with pytest.raises(NotImplementedError):
        pike_tmp_path.rglob("foo")


def test_iterdir(pike_tmp_path):
    filedata = {name: str(uuid.uuid4()) for name in ["f1", "f2", "f3", "f4"]}
    for f, data in filedata.items():
        (pike_tmp_path / f).write_text(data)
    for fpth in pike_tmp_path.iterdir():
        assert fpth.name in filedata
        assert fpth.read_text() == filedata[fpth.name]


def test_mkdir(pike_tmp_path):
    pth = pike_tmp_path / "foo" / "bar" / "baz"
    pth.mkdir(parents=True, exist_ok=False)
    pth.mkdir(exist_ok=True)
    while pth != pike_tmp_path:
        pth.rmdir()
        pth = pth.parent


def test_write_read(pike_tmp_path):
    test_bytes = b"hola pika" * (1024 * 1024)
    pth = pike_tmp_path / "bar.txt"
    pth.write_bytes(test_bytes)
    assert pth.read_bytes() == test_bytes


def test_open_w_r_a_x(pike_tmp_path):
    buf = "this is a test test test\nfoo\n\n"
    pth = pike_tmp_path / "foo.txt"
    with pth.open("w") as f:
        assert isinstance(f.buffer, io.BufferedWriter)
        assert f.writable()
        assert not f.readable()
        assert f.buffer.raw.create_request.create_disposition == smb2.FILE_SUPERSEDE
        f.write(buf)
    assert pth.read_bytes() == buf.encode("utf-8")
    with pth.open("r") as f:
        assert isinstance(f.buffer, io.BufferedReader)
        assert f.readable()
        assert not f.writable()
        assert f.buffer.raw.create_request.create_disposition == smb2.FILE_OPEN
        assert f.read() == buf
    with pth.open("a") as f:
        assert isinstance(f.buffer, io.BufferedWriter)
        assert f.writable()
        assert not f.readable()
        assert f.buffer.raw.create_request.create_disposition == smb2.FILE_OPEN_IF
        assert f.tell() == f.buffer.raw.end_of_file
        f.write("appended")
    assert pth.read_text().endswith("appended")
    with model.pike_status(ntstatus.STATUS_OBJECT_NAME_COLLISION):
        f = pth.open("x")
    pth.unlink()
    with pth.open("x") as f:
        f.write("exclusive access")
    assert pth.read_text() == "exclusive access"


@pytest.mark.parametrize("primary_mode", ("r", "w", "a", "rb", "wb", "ab"))
@pytest.mark.parametrize(
    "buffering", (pytest.param(-1, id="buffered"), pytest.param(0, id="raw"))
)
def test_open_plus(pike_tmp_path, primary_mode, buffering):
    buf = "this is a read/write test case\n"
    pth = pike_tmp_path / "foo.txt"
    if "r" in primary_mode:
        pth.touch()
    with pth.open(primary_mode + "+", buffering=buffering) as f:
        if "b" in primary_mode:
            buf = buf.encode("utf-8")
        if buffering == 0 and "b" in primary_mode:
            assert isinstance(f, pike.io.Open)
        elif "b" in primary_mode:
            assert isinstance(f, io.BufferedRandom)
        else:
            assert isinstance(f, io.TextIOWrapper)
            assert isinstance(f.buffer, io.BufferedRandom)
        assert f.readable()
        assert f.writable()
        f.write(buf)
        f.seek(0)
        assert f.read() == buf


def test_relative_to(pike_tmp_path):
    pth = pike_tmp_path / "foo"
    assert str(pth.relative_to(pike_tmp_path)) == "foo"


def test_rename(pike_tmp_path):
    orig_file = pike_tmp_path / "orig"
    new_file = pike_tmp_path / "new"

    orig_file.touch()
    assert orig_file.is_file()
    assert not new_file.exists()

    orig_file.rename(new_file)
    assert not orig_file.exists()
    assert new_file.is_file()


def test_parents(pike_tmp_path):
    pcs = ["foo", "bar", "baz"]
    pth = pike_tmp_path.joinpath(*pcs)
    parents = pth.parents
    for ix, pc in enumerate(reversed(pcs[:-1])):
        print(parents[ix])
        print(parents[ix].exists())
        assert parents[ix].name == pc


def test_join_from_root(pike_tmp_path, pike_TreeConnect):
    with pike_TreeConnect() as tc2:
        pth = pike_tmp_path / "rehome.txt"
        with pth.open("w") as f:
            f.write("foo the bar\n")
            # now attempt to open the file from the second connection
            with model.pike_status(ntstatus.STATUS_SHARING_VIOLATION):
                from_root = PikePath(tc2.tree).join_from_root(pth)
                print(from_root.read_text())


@pytest.mark.nosamba
def test_symlink(pike_tmp_path):
    buf = "this is my content\n"
    subdir = pike_tmp_path / "subdir"
    subdir.mkdir()
    target = subdir / "target"
    target.write_text(buf)
    link = pike_tmp_path / "link"
    link.symlink_to(target)
    assert link.read_text() == buf
    assert link.resolve() == target
    assert link.readlink()._path == target._path
    link.unlink()
    assert not link.exists()
    assert target.exists()


@pytest.mark.nosamba
def test_symlink_dir(pike_tmp_path):
    buf = "this is my content\n"
    subdir = pike_tmp_path / "subdir"
    subdir.mkdir()
    target = subdir / "target"
    target.write_text(buf)
    link = pike_tmp_path / "link"
    link.symlink_to(subdir)
    assert (link / target.name).read_text() == buf
    assert (link / target.name).resolve() == target
    assert link.resolve() == subdir
    assert link.readlink()._path == subdir._path


@pytest.mark.nosamba
def test_symlink_relative(pike_tmp_path):
    buf = "this is my content\n"
    subdir = pike_tmp_path / "subdir"
    subdir.mkdir()
    subdir_a2 = pike_tmp_path / "subdir2" / "a1" / "a2"
    subdir_a2.mkdir(parents=True)
    target = subdir / "target"
    target.write_text(buf)
    link = subdir_a2 / "link"
    link.symlink_to(target)
    assert link.read_text() == buf
    assert link.resolve() == target
    assert (
        link.readlink()._path == link.parent._path + "\\..\\..\\..\\..\\" + target._path
    )


@pytest.mark.nosamba
def test_symlink_chain(pike_tmp_path):
    buf = "this is my content\n"
    subdir = pike_tmp_path / "subdir"
    subdir.mkdir()
    subdir_a1 = pike_tmp_path / "subdir2" / "a1"
    subdir_a1.mkdir(parents=True)
    subdir_a2 = subdir_a1 / "a2"
    subdir_a2.mkdir()
    subdir_a3 = subdir_a2 / "a3"
    subdir_a3.mkdir()
    target = subdir / "target"
    target.write_text(buf)
    link = subdir_a3 / "link"
    link.symlink_to(subdir)
    link2 = subdir_a1 / "link2"
    link2.symlink_to(link)
    link = subdir_a2 / "link3"
    link.symlink_to(link2)
    link2 = subdir_a3 / "link4"
    link2.symlink_to(link)
    link = subdir_a1 / "link5"
    link.symlink_to(link2)
    assert (link / target.name).read_text() == buf
    assert (link / target.name).resolve() == target


def test_samefile(pike_tmp_path, pike_TreeConnect):
    f1 = pike_tmp_path / "f1"
    with pike_TreeConnect() as tc2:
        f1_2 = PikePath(tc2.tree).join_from_root(f1)
        assert f1_2.drive != f1.drive
        # when the file doesn't exist, it is not considered samefile
        assert not f1.samefile(f1_2)
        f1.touch()
        assert f1.samefile(f1_2)
        assert f1_2.samefile(f1)


def test_unlink(pike_tmp_path):
    fname = pike_tmp_path / str(uuid.uuid4())
    with model.pike_status(ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
        fname.unlink()
    fname.unlink(missing_ok=True)
    fname.touch()
    assert fname.is_file()
    fname.unlink()
    assert not fname.exists()
    fname.unlink(missing_ok=True)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        lease.py
#
# Abstract:
#
#        Lease tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

from builtins import map

import array
import random

import pike.model
import pike.smb2
import pike.test


@pike.test.RequireDialect(0x210)
@pike.test.RequireCapabilities(pike.smb2.SMB2_GLOBAL_CAP_LEASING)
class LeaseTest(pike.test.PikeTest):
    share_all = (
        pike.smb2.FILE_SHARE_READ
        | pike.smb2.FILE_SHARE_WRITE
        | pike.smb2.FILE_SHARE_DELETE
    )
    lease1 = array.array("B", map(random.randint, [0] * 16, [255] * 16))
    lease2 = array.array("B", map(random.randint, [0] * 16, [255] * 16))
    r = pike.smb2.SMB2_LEASE_READ_CACHING
    rw = r | pike.smb2.SMB2_LEASE_WRITE_CACHING
    rh = r | pike.smb2.SMB2_LEASE_HANDLE_CACHING
    rwh = rw | rh

    # Upgrade lease from RW to RWH, then break it to R
    def test_lease_upgrade_break(self):
        chan, tree = self.tree_connect()

        # Request rw lease
        handle1 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rw,
        ).result()

        self.assertEqual(handle1.lease.lease_state, self.rw)

        handle2 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
        ).result()

        self.assertIs(handle2.lease, handle1.lease)
        self.assertEqual(handle2.lease.lease_state, self.rwh)

        # On break, voluntarily give up handle caching
        handle2.lease.on_break(
            lambda state: state & ~pike.smb2.SMB2_LEASE_HANDLE_CACHING
        )

        # Break our lease
        handle3 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease2,
            lease_state=self.rwh,
        ).result()

        # First lease should have broken to r
        self.assertEqual(handle2.lease.lease_state, self.r)
        # Should be granted rh on second lease
        self.assertEqual(handle3.lease.lease_state, self.rh)

        chan.close(handle1)
        chan.close(handle2)
        chan.close(handle3)

    # Close handle associated with lease while a break is in progress
    def test_lease_break_close_ack(self):
        chan, tree = self.tree_connect()
        # Request rw lease
        handle1 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rw,
        ).result()

        # Upgrade to rwh
        handle2 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
        ).result()

        # Break our lease
        handle3_future = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease2,
            lease_state=self.rwh,
        )

        # Wait for break
        handle1.lease.future.wait()

        # Close second handle
        chan.close(handle2)

        # Now ack break
        handle1.lease.on_break(lambda state: state)

        # Wait for handle3
        handle3 = handle3_future.result()

        chan.close(handle1)
        chan.close(handle3)

    # Test that a lease can be shared across connections
    def test_lease_multiple_connections(self):
        chan, tree = self.tree_connect()

        # Request rwh lease
        handle1 = chan.create(
            tree,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
        ).result()
        self.assertEqual(handle1.lease.lease_state, self.rwh)

        chan2, tree2 = self.tree_connect()

        # Request rwh lease to same file on separate connection,
        # which we should get
        handle2 = chan2.create(
            tree2,
            "lease.txt",
            share=self.share_all,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
        ).result()
        self.assertEqual(handle2.lease.lease_state, self.rwh)

        # Leases should be the same object
        self.assertEqual(handle1.lease, handle2.lease)

        chan2.close(handle2)
        chan.close(handle1)
